import json
import mimetypes
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, cast

import httpx
from cognite.client.data_classes.data_modeling import ViewId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import (
    ErrorDetails,
    FailedResponse,
    HTTPClient,
    HTTPResult,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.http_client._item_classes import ItemsFailedResponse, ItemsResultList
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import NodeReference
from cognite_toolkit._cdf_tk.client.resource_classes.filemetadata import FileMetadataRequest, FileMetadataResponse
from cognite_toolkit._cdf_tk.cruds import FileMetadataCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.protocols import ResourceResponseProtocol
from cognite_toolkit._cdf_tk.utils import sanitize_filename
from cognite_toolkit._cdf_tk.utils.collection import chunker, chunker_sequence
from cognite_toolkit._cdf_tk.utils.fileio import MultiFileReader
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, UploadableStorageIO, UploadItem
from .selectors import FileContentSelector, FileIdentifierSelector, FileMetadataTemplateSelector
from .selectors._file_content import (
    FILEPATH,
    FileDataModelingTemplateSelector,
    FileExternalID,
    FileIdentifier,
    FileInstanceID,
    FileInternalID,
    FileTemplateSelector,
)
from .selectors._file_content import NodeId as SelectorNodeId

COGNITE_FILE_VIEW = ViewId("cdf_cdm", "CogniteFile", "v1")


class UploadFileContentItem(UploadItem[FileMetadataRequest]):
    file_path: Path
    mime_type: str

    def dump(self, camel_case: bool = True, exclude_extra: bool = True) -> dict[str, Any]:
        return self.item.dump(camel_case=camel_case, exclude_extra=exclude_extra)


@dataclass
class MetadataWithFilePath(ResourceResponseProtocol):
    metadata: FileMetadataResponse
    file_path: Path

    def as_write(self) -> FileMetadataRequest:
        return self.metadata.as_request_resource()


class FileContentIO(UploadableStorageIO[FileContentSelector, MetadataWithFilePath, FileMetadataRequest]):
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = FileContentSelector
    KIND = "FileContent"
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    UPLOAD_ENDPOINT = "/files"

    def __init__(self, client: ToolkitClient, target_dir: Path = Path.cwd()) -> None:
        super().__init__(client)
        self._crud = FileMetadataCRUD(client, None, None)
        self._target_dir = target_dir

    def as_id(self, item: MetadataWithFilePath) -> str:
        return item.metadata.external_id or str(item.metadata.id)

    def stream_data(
        self, selector: FileContentSelector, limit: int | None = None
    ) -> Iterable[Page[MetadataWithFilePath]]:
        if not isinstance(selector, FileIdentifierSelector):
            raise ToolkitNotImplementedError(
                f"Download with the manifest, {type(selector).__name__}, is not supported for FileContentIO"
            )
        selected_identifiers = selector.identifiers
        if limit is not None and limit < len(selected_identifiers):
            selected_identifiers = selected_identifiers[:limit]
        for identifiers in chunker_sequence(selected_identifiers, self.CHUNK_SIZE):
            metadata = self._retrieve_metadata(identifiers)
            if metadata is None:
                continue
            identifiers_map = self._as_metadata_map(metadata)
            downloaded_files: list[MetadataWithFilePath] = []
            for identifier in identifiers:
                if identifier not in identifiers_map:
                    continue

                meta = identifiers_map[identifier]
                filepath = self._create_filepath(meta, selector)
                download_url = self._retrieve_download_url(identifier)
                if download_url is None:
                    continue
                filepath.parent.mkdir(parents=True, exist_ok=True)
                with httpx.stream("GET", download_url) as response:
                    if response.status_code != 200:
                        continue
                    with filepath.open(mode="wb") as file_stream:
                        for chunk in response.iter_bytes(chunk_size=8192):
                            file_stream.write(chunk)
                downloaded_files.append(
                    MetadataWithFilePath(
                        metadata=meta,
                        file_path=filepath.relative_to(self._target_dir),
                    )
                )
            yield Page(items=downloaded_files, worker_id="Main")

    def _retrieve_metadata(self, identifiers: Sequence[FileIdentifier]) -> Sequence[FileMetadataResponse] | None:
        config = self.client.config
        response = self.client.http_client.request_single_retries(
            message=RequestMessage(
                endpoint_url=config.create_api_url("/files/byids"),
                method="POST",
                body_content={
                    "items": [
                        identifier.model_dump(mode="json", by_alias=True, exclude={"id_type"})
                        for identifier in identifiers
                    ],
                    "ignoreUnknownIds": True,
                },
            )
        )
        if not isinstance(response, SuccessResponse):
            return None
        try:
            body = response.body_json
        except ValueError:
            return None

        items_data = body.get("items", [])
        if not isinstance(items_data, list):
            return None
        return [FileMetadataResponse.model_validate(item) for item in items_data]

    @staticmethod
    def _as_metadata_map(metadata: Sequence[FileMetadataResponse]) -> dict[FileIdentifier, FileMetadataResponse]:
        identifiers_map: dict[FileIdentifier, FileMetadataResponse] = {}
        for item in metadata:
            if item.id is not None:
                identifiers_map[FileInternalID(internal_id=item.id)] = item
            if item.external_id is not None:
                identifiers_map[FileExternalID(external_id=item.external_id)] = item
            if item.instance_id is not None:
                identifiers_map[
                    FileInstanceID(
                        instance_id=SelectorNodeId(
                            space=item.instance_id.space, external_id=item.instance_id.external_id
                        )
                    )
                ] = item
        return identifiers_map

    def _create_filepath(self, meta: FileMetadataResponse, selector: FileIdentifierSelector) -> Path:
        # We now that metadata always have name set
        filename = Path(sanitize_filename(meta.name))
        if len(filename.suffix) == 0 and meta.mime_type:
            if mime_ext := mimetypes.guess_extension(meta.mime_type):
                filename = filename.with_suffix(mime_ext)
        directory = sanitize_filename(selector.file_directory)
        if isinstance(meta.directory, str) and meta.directory != "":
            directory = sanitize_filename(meta.directory.removeprefix("/"))

        counter = 1
        filepath = self._target_dir / directory / filename
        while filepath.exists():
            filepath = self._target_dir / directory / f"{filename} ({counter})"
            counter += 1

        return filepath

    def _retrieve_download_url(self, identifier: FileIdentifier) -> str | None:
        config = self.client.config
        response = self.client.http_client.request_single_retries(
            message=RequestMessage(
                endpoint_url=config.create_api_url("/files/downloadlink"),
                method="POST",
                body_content={"items": [identifier.model_dump(mode="json", by_alias=True, exclude={"id_type"})]},
            )
        )
        if not isinstance(response, SuccessResponse):
            return None

        try:
            body = response.body_json
        except ValueError:
            return None

        if "items" in body and isinstance(body["items"], list) and len(body["items"]) > 0:
            # The API responses is not following the API docs, this is a workaround
            body = body["items"][0]
        try:
            return cast(str, body["downloadUrl"])
        except (KeyError, IndexError):
            return None

    def count(self, selector: FileContentSelector) -> int | None:
        if isinstance(selector, FileIdentifierSelector):
            return len(selector.identifiers)
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[MetadataWithFilePath], selector: FileContentSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        """Convert a writable Cognite resource list to a JSON-compatible chunk of data.

        Args:
            data_chunk: A writable Cognite resource list representing the data.
            selector: The selector used for the data. (Not used in this implementation)
        Returns:
            A list of dictionaries, each representing the data in a JSON-compatible format.
        """
        result: list[dict[str, JsonVal]] = []
        for item in data_chunk:
            item_json = self._crud.dump_resource(item.metadata)
            item_json[FILEPATH] = item.file_path.as_posix()
            result.append(item_json)
        return result

    def json_chunk_to_data(self, data_chunk: list[tuple[str, dict[str, JsonVal]]]) -> Sequence[UploadFileContentItem]:
        """Convert a JSON-compatible chunk of data back to a writable Cognite resource list.

        Args:
            data_chunk: A list of tuples, each containing a source ID and a dictionary representing
                the data in a JSON-compatible format.
        Returns:
            A writable Cognite resource list representing the data.
        """
        result: list[UploadFileContentItem] = []
        for source_id, item_json in data_chunk:
            item = self.json_to_resource(item_json)
            filepath = Path(cast(str | Path, item_json[FILEPATH]))
            mime_type, _ = mimetypes.guess_type(filepath)
            # application/octet-stream is the standard fallback for binary data when the type is unknown. (at least Claude thinks so)
            result.append(
                UploadFileContentItem(
                    source_id=source_id,
                    item=item,
                    file_path=filepath,
                    mime_type=mime_type or "application/octet-stream",
                )
            )
        return result

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> FileMetadataRequest:
        return self._crud.load_resource(item_json)

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[FileMetadataRequest]],
        http_client: HTTPClient,
        selector: FileContentSelector | None = None,
    ) -> ItemsResultList:
        results = ItemsResultList()
        if isinstance(selector, FileMetadataTemplateSelector | FileIdentifierSelector):
            upload_url_getter = self._upload_url_asset_centric
        elif isinstance(selector, FileDataModelingTemplateSelector):
            upload_url_getter = self._upload_url_data_modeling
        elif selector is None:
            raise ToolkitNotImplementedError("Selector must be provided for FileContentIO upload")
        else:
            raise ToolkitNotImplementedError(
                f"Upload for the given selector, {type(selector).__name__}, is not supported for FileContentIO"
            )

        for item in cast(Sequence[UploadFileContentItem], data_chunk):
            if not (upload_url := upload_url_getter(item, http_client, results)):
                continue

            content_bytes = item.file_path.read_bytes()
            upload_response = http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=upload_url,
                    method="PUT",
                    content_type=item.mime_type,
                    data_content=content_bytes,
                )
            )
            results.append(upload_response.as_item_response(str(item)))
        return results

    def _upload_url_asset_centric(
        self, item: UploadFileContentItem, http_client: HTTPClient, results: ItemsResultList
    ) -> str | None:
        response = http_client.request_single_retries(
            message=RequestMessage(
                endpoint_url=http_client.config.create_api_url(self.UPLOAD_ENDPOINT),
                method="POST",
                body_content=item.dump(),
            )
        )
        return self._parse_upload_link_response(response, item, results)

    def _upload_url_data_modeling(
        self,
        item: UploadFileContentItem,
        http_client: HTTPClient,
        results: ItemsResultList,
        created_node: bool = False,
    ) -> str | None:
        """Get upload URL for data modeling file upload.

        We first try to get the upload link assuming the CogniteFile node already exists.
        If we get a "not found" error, we create the CogniteFile node and try again.

        Args:
            item: The upload item containing file metadata.
            http_client: The HTTP client to use for requests.
            results: A mutable sequence to collect HTTP messages and errors.
            created_node: A flag indicating whether the CogniteFile node has already been created.
                This prevents infinite recursion.

        Returns:
            The upload URL as a string, or None if there was an error.

        """
        # We know that instance_id is always set for data modeling uploads
        instance_id = cast(NodeReference, item.item.instance_id)
        response = http_client.request_single_retries(
            message=RequestMessage(
                endpoint_url=http_client.config.create_api_url("/files/uploadlink"),
                method="POST",
                body_content={"items": [{"instanceId": instance_id.dump()}]},
            )
        )
        if isinstance(response, FailedResponse) and response.error.missing and not created_node:
            if self._create_cognite_file_node(instance_id, http_client, item.source_id, results):
                return self._upload_url_data_modeling(item, http_client, results, created_node=True)
            else:
                return None

        return self._parse_upload_link_response(response, item, results)

    @classmethod
    def _create_cognite_file_node(
        cls, instance_id: NodeReference, http_client: HTTPClient, upload_id: str, results: ItemsResultList
    ) -> bool:
        node_creation = http_client.request_single_retries(
            message=RequestMessage(
                endpoint_url=http_client.config.create_api_url("/models/instances"),
                method="POST",
                body_content={
                    "items": [
                        {
                            "space": instance_id.space,
                            "externalId": instance_id.external_id,
                            "instanceType": "node",
                            # When we create a node with properties in CogniteFile View even with empty properties,
                            # CDF will fill in empty values for all properties defined in the view (note this is only
                            # possible because CogniteFile view has all properties as optional). This includes properties
                            # in the CogniteFile container, which will trigger the file syncer to create a FileMetadata
                            # and link it to the CogniteFile node.
                            "sources": [{"source": COGNITE_FILE_VIEW.dump(include_type=True), "properties": {}}],  # type: ignore[dict-item]
                        }
                    ]
                },
            )
        )
        if isinstance(node_creation, SuccessResponse):
            # Node created successfully
            return True
        results.append(node_creation.as_item_response(upload_id))
        return False

    @classmethod
    def _parse_upload_link_response(
        cls, response: HTTPResult, item: UploadFileContentItem, results: ItemsResultList
    ) -> str | None:
        if not isinstance(response, SuccessResponse):
            results.append(response.as_item_response(item.source_id))
            return None
        try:
            body = response.body_json
        except ValueError:
            results.append(
                ItemsFailedResponse(
                    status_code=response.status_code,
                    body=response.body,
                    error=ErrorDetails(code=response.status_code, message="Invalid JSON response"),
                    ids=[item.source_id],
                )
            )
            return None
        if "items" in body and isinstance(body["items"], list) and len(body["items"]) > 0:
            body = body["items"][0]
        try:
            upload_url = cast(str, body["uploadUrl"])
        except (KeyError, IndexError):
            results.append(
                ItemsFailedResponse(
                    status_code=200,
                    body=json.dumps(body),
                    error=ErrorDetails(code=200, message="Malformed response"),
                    ids=[item.source_id],
                )
            )
            return None
        return upload_url

    @classmethod
    def read_chunks(
        cls, reader: MultiFileReader, selector: FileContentSelector
    ) -> Iterable[list[tuple[str, dict[str, JsonVal]]]]:
        if isinstance(selector, FileTemplateSelector):
            for chunk in chunker_sequence(reader.input_files, cls.CHUNK_SIZE):
                batch: list[tuple[str, dict[str, JsonVal]]] = []
                for file_path in chunk:
                    metadata = selector.create_instance(file_path)
                    metadata[FILEPATH] = file_path
                    batch.append((file_path.as_posix(), metadata))
                yield batch
        elif isinstance(selector, FileIdentifierSelector):
            for item_chunk in chunker(reader.read_chunks(), cls.CHUNK_SIZE):
                batch = []
                for item in item_chunk:
                    if FILEPATH not in item:
                        # Todo Log warning
                        continue
                    try:
                        file_path = Path(item[FILEPATH])
                    except KeyError:
                        # Todo Log warning
                        continue
                    if not file_path.is_absolute():
                        file_path = reader.input_file.parent / file_path
                    item[FILEPATH] = file_path
                    batch.append((file_path.as_posix(), item))
                yield batch
        else:
            raise ToolkitNotImplementedError(
                f"Reading with the manifest, {type(selector).__name__}, is not supported for FileContentIO"
            )

    @classmethod
    def count_items(cls, reader: MultiFileReader, selector: FileContentSelector | None = None) -> int:
        # We upload one file at a time, so the number of items is the number of input files
        return len(reader.input_files)



from functools import lru_cache

from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, RequestMessage
from cognite_toolkit._cdf_tk.client.resource_classes.project import OrganizationResponse, ProjectStatusList


class ProjectAPI:
    def __init__(self, http_client: HTTPClient) -> None:
        self._http_client = http_client

    def status(self) -> ProjectStatusList:
        """Retrieve information about the current project."""
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=f"{self._http_client.config.base_url}/api/v1/projects",
                method="GET",
                parameters={"withDataModelingStatus": True},
            )
        ).get_success_or_raise()
        result = ProjectStatusList.model_validate_json(response.body)
        result._project = self._http_client.config.project
        return result

    def organization(self) -> OrganizationResponse:
        """Retrieve information about the organization of the current project."""
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._http_client.config.base_api_url,
                method="GET",
            )
        )
        success = response.get_success_or_raise()
        return OrganizationResponse.model_validate_json(success.body)

    @lru_cache(maxsize=1)
    def get_organization_id(self) -> str:
        """Retrieve the organization id of the current project."""
        return self.organization().organization



from collections.abc import Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId
from cognite_toolkit._cdf_tk.client.resource_classes.location_filter import (
    LocationFilterRequest,
    LocationFilterResponse,
)


class LocationFiltersAPI(CDFResourceAPI[LocationFilterResponse]):
    """API for managing Location Filters using the CDFResourceAPI pattern.

    This API manages location filter configurations at:
    /apps/v1/projects/{project}/storage/config/locationfilters
    """

    BASE_PATH = "/storage/config/locationfilters"

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path=self.BASE_PATH, item_limit=1),
                "retrieve": Endpoint(method="POST", path=f"{self.BASE_PATH}/byids", item_limit=1000),
                "update": Endpoint(method="PUT", path=f"{self.BASE_PATH}/{{id}}", item_limit=1),
                "delete": Endpoint(method="DELETE", path=f"{self.BASE_PATH}/{{id}}", item_limit=1),
                "list": Endpoint(method="POST", path=f"{self.BASE_PATH}/list", item_limit=1000),
            },
        )

    def _make_url(self, path: str = "") -> str:
        """Create the full URL for this resource endpoint using the apps URL format."""
        return self._http_client.config.create_app_url(path)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[LocationFilterResponse]:
        return PagedResponse[LocationFilterResponse].model_validate_json(response.body)

    def create(self, items: Sequence[LocationFilterRequest]) -> list[LocationFilterResponse]:
        """Create a new location filter.

        Args:
            items: The location filter to create.

        Returns:
            The created location filter.
        """
        endpoint = self._method_endpoint_map["create"]
        results: list[LocationFilterResponse] = []
        for item in items:
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body_content=item.model_dump(mode="json", by_alias=True),
            )
            result = self._http_client.request_single_retries(request)
            response = result.get_success_or_raise()
            results.append(LocationFilterResponse.model_validate_json(response.body))
        return results

    def retrieve(self, items: Sequence[InternalId]) -> list[LocationFilterResponse]:
        """Retrieve a single location filter by ID.

        Args:
            items: The ID of the location filter to retrieve.

        Returns:
            The retrieved location filter.
        """
        return self._request_item_response(items, "retrieve")

    def update(self, items: Sequence[LocationFilterRequest]) -> list[LocationFilterResponse]:
        """Update an existing location filter.

        Args:
            items: The updated location filter data.

        Returns:
            The updated location filter.
        """
        endpoint = self._method_endpoint_map["update"]
        results: list[LocationFilterResponse] = []
        for item in items:
            if item.id is None:
                raise ValueError("Item must have an ID for update operation.")
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path.format(id=item.id)),
                method=endpoint.method,
                body_content=item.model_dump(mode="json", by_alias=True),
            )
            result = self._http_client.request_single_retries(request)
            response = result.get_success_or_raise()
            parsed = LocationFilterResponse.model_validate_json(response.body)
            parsed.id = item.id
            results.append(parsed)
        return results

    def delete(self, items: Sequence[InternalId]) -> list[LocationFilterResponse]:
        """Delete a location filter.

        Args:
            items: The ID of the location filter to delete.

        Returns:
            list[LocationFilterResponse]: The deleted location filters.
        """
        endpoint = self._method_endpoint_map["delete"]
        results: list[LocationFilterResponse] = []
        for item in items:
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path.format(id=item.id)),
                method=endpoint.method,
            )
            result = self._http_client.request_single_retries(request)
            response = result.get_success_or_raise()
            results.append(LocationFilterResponse.model_validate_json(response.body))
        return results

    # Overwritten to avoid passing limit to the body
    def _paginate(
        self,
        limit: int,
        cursor: str | None = None,
        params: dict[str, Any] | None = None,
        body: dict[str, Any] | None = None,
        endpoint_path: str | None = None,
    ) -> PagedResponse[LocationFilterResponse]:
        endpoint = self._method_endpoint_map["list"]
        request = RequestMessage(
            endpoint_url=self._make_url(endpoint.path),
            method=endpoint.method,
            body_content=body or {},
        )
        result = self._http_client.request_single_retries(request)
        response = result.get_success_or_raise()
        return self._validate_page_response(response)

    def paginate(
        self,
        flat: bool = True,
    ) -> PagedResponse[LocationFilterResponse]:
        """Get a single page of location filters.

        Args:
            flat: Whether to return a flat list (default True).

        Returns:
            PagedResponse of LocationFilterResponse objects.
        """
        return self._paginate(cursor=None, limit=100, body={"flat": flat})

    def iterate(self, flat: bool = True) -> Iterable[list[LocationFilterResponse]]:
        """Iterate over all location filters.

        Args:
            flat: Whether to return a flat list (default True).

        Returns:
            Iterable of lists of LocationFilterResponse objects.
        """
        return self._iterate(limit=None, body={"flat": flat})

    def list(self, flat: bool = True) -> list[LocationFilterResponse]:
        """List all location filters.

        Args:
            flat: Whether to return a flat list (default True).

        Returns:
            List of LocationFilterResponse objects.
        """
        return self._list(limit=None, body={"flat": flat})



from dataclasses import dataclass
from enum import Enum
from functools import lru_cache

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import clean_name


@dataclass(frozen=True)
class FlagMetadata:
    """Metadata for a feature flag.

    Attributes:
        visible: If True, the flag is shown in 'cdf about' even when disabled,
            allowing users to discover it. If False, the flag is hidden unless
            explicitly enabled in cdf.toml.
        description: Human-readable description of what the flag enables.
    """

    visible: bool
    description: str


class Flags(Enum):
    INTERNAL = FlagMetadata(
        visible=False,
        description="Does nothing",
    )
    IMPORT_CMD = FlagMetadata(
        visible=True,
        description="Enables the import sub application",
    )
    GRAPHQL = FlagMetadata(
        visible=False,
        description="Enables the support for deploying data models as GraphQL schemas",
    )
    MODULE_REPEAT = FlagMetadata(
        visible=True,
        description="Enables the support for repeating modules in the config file",
    )
    PROFILE = FlagMetadata(
        visible=True,
        description="Enables support for the profile command",
    )
    INFIELD = FlagMetadata(
        visible=True,
        description="Enables support for Infield configs",
    )
    MIGRATE = FlagMetadata(
        visible=True,
        description="Enables the migrate command",
    )
    INFIELD_MIGRATE = FlagMetadata(
        visible=False,
        description="Enables the infield-migrate command for migrating Infield configs to regular Toolkit configs",
    )
    STREAMS = FlagMetadata(
        visible=True,
        description="Enables the support for the streams resources",
    )
    v08 = FlagMetadata(
        visible=False,
        description="Enables features planned for Cognite Toolkit version 0.8.0",
    )
    CREATE = FlagMetadata(
        visible=True,
        description="Enables the support for the resources create command under dev plugin",
    )
    EXTEND_DOWNLOAD = FlagMetadata(
        visible=True,
        description="Enables extended download to support downloading file content and datapoints",
    )
    EXTEND_UPLOAD = FlagMetadata(
        visible=True,
        description="Enables extended upload to support uploading individual files and records",
    )
    SIMULATORS = FlagMetadata(
        visible=True,
        description="Enables the support for simulator model resources",
    )
    FUNCTION_REQUIREMENTS_VALIDATION = FlagMetadata(
        visible=True,
        description="Enables validation of function requirements.txt during build using pip dry-run",
    )
    RESPACE = FlagMetadata(
        visible=False,
        description="Enables the respace command for moving nodes between spaces",
    )
    DATA_PRODUCTS = FlagMetadata(
        visible=False,
        description="Enables the support for data product resources",
    )
    MODULES_LIST_JSON = FlagMetadata(
        visible=True,
        description="Enables JSON output format for the modules list command",
    )

    def is_enabled(self) -> bool:
        return FeatureFlag.is_enabled(self)


class FeatureFlag:
    @staticmethod
    @lru_cache(typed=True)
    def is_enabled(flag: Flags) -> bool:
        return CDFToml.load().alpha_flags.get(clean_name(flag.name), False)

    @staticmethod
    def flush() -> None:
        FeatureFlag.is_enabled.cache_clear()



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections.abc import Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any, final

import yaml
from cognite.client.data_classes.capabilities import (
    Capability,
    ExtractionPipelinesAcl,
)

from cognite_toolkit._cdf_tk.client.http_client import ToolkitAPIError
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.extraction_pipeline import (
    ExtractionPipelineRequest,
    ExtractionPipelineResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.extraction_pipeline_config import (
    ExtractionPipelineConfigRequest,
    ExtractionPipelineConfigResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import (
    ExternalId,
    ExtractionPipelineConfigId,
    InternalOrExternalId,
    RawDatabaseId,
    RawTableId,
)
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.resource_classes import ExtractionPipelineConfigYAML, ExtractionPipelineYAML
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
)
from cognite_toolkit._cdf_tk.utils import (
    load_yaml_inject_variables,
    read_yaml_content,
    safe_read,
    sanitize_filename,
    stringify_value_by_key_in_yaml,
)
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_force_hashable, diff_list_identifiable

from .auth import GroupAllScopedCRUD
from .data_organization import DataSetsCRUD
from .raw import RawDatabaseCRUD, RawTableCRUD


@final
class ExtractionPipelineCRUD(ResourceCRUD[ExternalId, ExtractionPipelineRequest, ExtractionPipelineResponse]):
    folder_name = "extraction_pipelines"
    resource_cls = ExtractionPipelineResponse
    resource_write_cls = ExtractionPipelineRequest
    kind = "ExtractionPipeline"
    dependencies = frozenset({DataSetsCRUD, RawDatabaseCRUD, RawTableCRUD, GroupAllScopedCRUD})
    yaml_cls = ExtractionPipelineYAML
    _doc_url = "Extraction-Pipelines/operation/createExtPipes"

    @property
    def display_name(self) -> str:
        return "extraction pipelines"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ExtractionPipelineRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [ExtractionPipelinesAcl.Action.Read]
            if read_only
            else [ExtractionPipelinesAcl.Action.Read, ExtractionPipelinesAcl.Action.Write]
        )

        scope: ExtractionPipelinesAcl.Scope.All | ExtractionPipelinesAcl.Scope.DataSet = (  # type: ignore[valid-type]
            ExtractionPipelinesAcl.Scope.All()
        )
        if items is not None:
            if data_set_id := {item.data_set_id for item in items if item.data_set_id}:
                scope = ExtractionPipelinesAcl.Scope.DataSet(list(data_set_id))

        return ExtractionPipelinesAcl(actions, scope)

    @classmethod
    def get_id(cls, item: ExtractionPipelineRequest | ExtractionPipelineResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise ToolkitRequiredValueError("ExtractionPipeline must have external_id set.")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: ExtractionPipelineResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if item.id is None:
            raise ToolkitRequiredValueError("ExtractionPipeline must have id set.")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return sanitize_filename(id.external_id)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        seen_databases: set[str] = set()
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])
        if "rawTables" in item:
            for entry in item["rawTables"]:
                if db := entry.get("dbName"):
                    if db not in seen_databases:
                        seen_databases.add(db)
                        yield RawDatabaseCRUD, RawDatabaseId(name=db)
                    if "tableName" in entry:
                        yield RawTableCRUD, RawTableId(db_name=db, name=entry["tableName"])

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> ExtractionPipelineRequest:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if "createdBy" not in resource:
            # Todo; Bug SDK missing default value (this will be set on the server-side if missing)
            resource["createdBy"] = "unknown"
        return ExtractionPipelineRequest.model_validate(resource)

    def dump_resource(
        self, resource: ExtractionPipelineResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if dumped.get("createdBy") == "unknown" and "createdBy" not in local:
            dumped.pop("createdBy", None)
        elif dumped.get("createdBy") == "unknown" and "createdBy" in local and local["createdBy"] is None:
            dumped["createdBy"] = None
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("rawTables",):
            return diff_list_identifiable(local, cdf, get_identifier=lambda x: (x["dbName"], x["tableName"]))
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[ExtractionPipelineRequest]) -> list[ExtractionPipelineResponse]:
        return self.client.tool.extraction_pipelines.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[ExtractionPipelineResponse]:
        return self.client.tool.extraction_pipelines.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[ExtractionPipelineRequest]) -> list[ExtractionPipelineResponse]:
        return self.client.tool.extraction_pipelines.update(list(items), mode="replace")

    def delete(self, ids: Sequence[InternalOrExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.extraction_pipelines.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[ExtractionPipelineResponse]:
        filter: ClassicFilter | None = None
        if data_set_external_id is not None:
            filter = ClassicFilter(data_set_ids=[ExternalId(external_id=data_set_external_id)])
        for pipelines in self.client.tool.extraction_pipelines.iterate(filter=filter, limit=None):
            yield from pipelines


@final
class ExtractionPipelineConfigCRUD(
    ResourceCRUD[
        ExternalId,
        ExtractionPipelineConfigRequest,
        ExtractionPipelineConfigResponse,
    ]
):
    folder_name = "extraction_pipelines"
    resource_cls = ExtractionPipelineConfigResponse
    resource_write_cls = ExtractionPipelineConfigRequest
    kind = "Config"
    dependencies = frozenset({ExtractionPipelineCRUD})
    _doc_url = "Extraction-Pipelines-Config/operation/createExtPipeConfig"
    parent_resource = frozenset({ExtractionPipelineCRUD})
    yaml_cls = ExtractionPipelineConfigYAML

    support_update = False

    @property
    def display_name(self) -> str:
        return "extraction pipeline configs"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ExtractionPipelineConfigRequest] | None, read_only: bool
    ) -> list[Capability]:
        # We check the parent extraction pipeline permissions instead
        return []

    @classmethod
    def get_id(cls, item: ExtractionPipelineConfigRequest | ExtractionPipelineConfigResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise ToolkitRequiredValueError("ExtractionPipelineConfig must have external_id set.")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return sanitize_filename(id.external_id)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "externalId" in item:
            yield ExtractionPipelineCRUD, ExternalId(external_id=item["externalId"])

    def safe_read(self, filepath: Path | str) -> str:
        # The config is expected to be a string that is parsed as a YAML on the server side.
        # The user typically writes the config as an object, so add a | to ensure it is parsed as a string.
        return stringify_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="config")

    def load_resource_file(
        # special case where the environment variable keys in the 'config' value
        # should not be replaced but preserved as is
        self,
        filepath: Path,
        environment_variables: dict[str, str | None] | None = None,
    ) -> list[dict[str, Any]]:
        raw_str = self.safe_read(filepath)

        original = load_yaml_inject_variables(raw_str, {}, validate=False, original_filepath=filepath)
        replaced = load_yaml_inject_variables(
            raw_str, environment_variables or {}, validate=False, original_filepath=filepath
        )

        if isinstance(original, dict) and isinstance(replaced, dict):
            if "config" in original:
                replaced["config"] = original.get("config")
                return [replaced]

        for orig_item, repl_item in zip(original, replaced):
            if "config" in orig_item:
                repl_item["config"] = orig_item.get("config")  # type: ignore
        return replaced if isinstance(replaced, list) else [replaced]

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> ExtractionPipelineConfigRequest:
        config_raw = resource.get("config")
        if isinstance(config_raw, str):
            self._validate_config(config_raw, resource)
        return ExtractionPipelineConfigRequest.model_validate(resource)

    def _validate_config(self, config_raw: str, resource: dict[str, Any]) -> dict[str, Any]:
        """The extraction pipeline API specifies config to be a string. However, it should be a valid YAML mapping,
        so we validate that here and warn if it is not."""
        try:
            result = read_yaml_content(config_raw)
        except yaml.YAMLError as e:
            id_ = self._get_id(resource, default="missing")
            HighSeverityWarning(
                f"Configuration for {id_!r} could not be parsed "
                f"as valid YAML, which is the recommended format. Error: {e}"
            ).print_warning(console=self.console)
        else:
            if isinstance(result, dict):
                return result
            id_ = self._get_id(resource, default="missing")
            HighSeverityWarning(
                f"Configuration for {id_!r} is not a valid YAML mapping (dict). Got {type(result).__name__} instead."
            ).print_warning(console=self.console)
        return {}

    def _get_id(self, resource: dict[str, Any], default: str) -> str:
        try:
            return str(self.get_id(resource))
        except (ToolkitRequiredValueError, KeyError):
            return default

    def dump_resource(
        self, resource: ExtractionPipelineConfigResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if (
            "config" in dumped
            and isinstance(dumped["config"], str)
            and ("config" not in local or isinstance(local["config"], dict))
        ):
            # When we dump a config from CDF, i.e., running `cdf dump extraction-pipeline`, we want to parse the config
            # as YAML to make it easier to read and edit.
            if dumped["config"].strip() == "":
                dumped["config"] = {}
            else:
                dumped["config"] = self._validate_config(dumped["config"], dumped)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "config":
            # Assume all arrays in the config are hashable
            return diff_list_force_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)

    def _upsert(self, items: Sequence[ExtractionPipelineConfigRequest]) -> list[ExtractionPipelineConfigResponse]:
        return self.client.tool.extraction_pipelines.configs.create(list(items))

    def create(self, items: Sequence[ExtractionPipelineConfigRequest]) -> list[ExtractionPipelineConfigResponse]:
        return self._upsert(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[ExtractionPipelineConfigResponse]:
        return self.client.tool.extraction_pipelines.configs.retrieve(
            [ExtractionPipelineConfigId(external_id=pipeline_id.external_id) for pipeline_id in ids]
        )

    def delete(self, ids: Sequence[ExternalId]) -> int:
        """Delete is not supported for extraction pipeline configs.

        Instead, we assume that when the user deletes the extraction pipeline configs, they are also deleting the
        extraction pipelines which will automatically delete the configs. In this method, we simply count the number
        of configs that exist for the given ids and return that number as these will be deleted.
        """
        # Todo; Change to raise ToolkitNotSupportedError in v0.8
        count = 0
        for id_ in ids:
            try:
                result = self.client.tool.extraction_pipelines.configs.list(external_id=id_.external_id, limit=None)
            except ToolkitAPIError as e:
                if e.code == 403 and "not found" in e.message and "extraction pipeline" in e.message.lower():
                    continue
                raise
            else:
                count += len(result)
        return count

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[ExtractionPipelineConfigResponse]:
        if parent_ids is None:
            parent_external_ids: Iterable[ExternalId] = (
                pipeline.as_id() for pipeline in self.client.tool.extraction_pipelines.list(limit=None)
            )
        else:
            parent_external_ids = [pid for pid in parent_ids if isinstance(pid, ExternalId)]
        for parent_id in parent_external_ids:
            try:
                for configs in self.client.tool.extraction_pipelines.configs.iterate(
                    external_id=parent_id.external_id, limit=None
                ):
                    yield from configs
            except ToolkitAPIError as e:
                if e.code == 404 and "There is no config stored" in e.message:
                    continue
                raise



from typing import Any, ClassVar, Literal

from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, ResponseResource, UpdatableRequestResource
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId
from .instance_api import NodeReference


class FileMetadata(BaseModelObject):
    external_id: str | None = None
    name: str
    directory: str | None = None
    source: str | None = None
    mime_type: str | None = None
    metadata: Metadata | None = None
    asset_ids: list[int] | None = None
    data_set_id: int | None = None
    labels: list[dict[Literal["externalId"], str]] | None = None
    geo_location: JsonValue | None = None
    source_created_time: int | None = None
    source_modified_time: int | None = None
    security_categories: list[int] | None = None

    def as_id(self) -> ExternalId:
        if self.external_id is None:
            raise ValueError("Cannot convert FileMetadataRequest to ExternalId when external_id is None")
        return ExternalId(external_id=self.external_id)


class FileMetadataRequest(FileMetadata, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata", "labels", "asset_ids", "security_categories"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"asset_ids", "security_categories"})
    # This field is not part of the request when creating or updating a resource
    # but we added it here for convenience so that it is available when converting
    # from response to request.
    instance_id: NodeReference | None = Field(default=None, exclude=True)

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        update = super().as_update(mode)
        # Name cannot be updated.
        update["update"].pop("name", None)
        return update


class FileMetadataResponse(FileMetadata, ResponseResource[FileMetadataRequest]):
    created_time: int
    last_updated_time: int
    uploaded_time: int | None = None
    uploaded: bool
    id: int
    instance_id: NodeReference | None = None
    pending_instance_id: NodeReference | None = None
    # This field is required in the upload endpoint response, but not in any other file metadata response
    upload_url: str | None = None

    @classmethod
    def request_cls(cls) -> type[FileMetadataRequest]:
        return FileMetadataRequest



"""Function Schedules API for managing CDF function schedules.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Function-schedules/operation/postFunctionSchedules
"""

from collections.abc import Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.function_schedule import (
    FunctionScheduleRequest,
    FunctionScheduleResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId


class FunctionSchedulesAPI(CDFResourceAPI[FunctionScheduleResponse]):
    """API for managing CDF function schedules.

    Note: Function schedules do not support update operations.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/functions/schedules", item_limit=1),
                "retrieve": Endpoint(method="POST", path="/functions/schedules/byids", item_limit=10_000),
                "delete": Endpoint(method="POST", path="/functions/schedules/delete", item_limit=10_000),
                "list": Endpoint(method="POST", path="/functions/schedules/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[FunctionScheduleResponse]:
        return PagedResponse[FunctionScheduleResponse].model_validate_json(response.body)

    def create(self, items: Sequence[FunctionScheduleRequest]) -> list[FunctionScheduleResponse]:
        """Create function schedules in CDF.

        Args:
            items: List of FunctionScheduleRequest objects to create.

        Returns:
            List of created FunctionScheduleResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(self, items: Sequence[InternalId], ignore_unknown_ids: bool = False) -> list[FunctionScheduleResponse]:
        """Retrieve function schedules from CDF by ID.

        Args:
            items: List of InternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.

        Returns:
            List of retrieved FunctionScheduleResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def delete(self, items: Sequence[InternalId]) -> None:
        """Delete function schedules from CDF.

        Args:
            items: List of InternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def _create_list_request_body(self, function_id: int | None) -> dict[str, Any] | None:
        body: dict[str, Any] = {}
        if function_id is not None:
            body["filter"] = {"functionId": function_id}
        return body or None

    def paginate(
        self,
        function_id: int | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[FunctionScheduleResponse]:
        """Get a page of function schedules from CDF.

        Args:
            limit: Maximum number of function schedules to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of FunctionScheduleResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body=self._create_list_request_body(function_id=function_id),
        )

    def iterate(
        self,
        function_id: int | None = None,
        limit: int | None = None,
    ) -> Iterable[list[FunctionScheduleResponse]]:
        """Iterate over all function schedules in CDF.

        Args:
            limit: Maximum total number of function schedules to return.

        Returns:
            Iterable of lists of FunctionScheduleResponse objects.
        """
        return self._iterate(limit=limit, body=self._create_list_request_body(function_id=function_id))

    def list(self, function_id: int | None = None, limit: int | None = None) -> list[FunctionScheduleResponse]:
        """List all function schedules in CDF.

        Args:
            limit: Maximum total number of function schedules to return.

        Returns:
            List of FunctionScheduleResponse objects.
        """
        return self._list(limit=limit, body=self._create_list_request_body(function_id=function_id))



from collections.abc import Iterable, Mapping, Sequence
from itertools import groupby
from typing import Any, ClassVar, cast

from cognite.client._proto.data_point_insertion_request_pb2 import DataPointInsertionItem, DataPointInsertionRequest
from cognite.client._proto.data_point_list_response_pb2 import DataPointListResponse
from cognite.client._proto.data_points_pb2 import (
    NumericDatapoint,
    NumericDatapoints,
    StringDatapoint,
    StringDatapoints,
)
from cognite.client.data_classes import TimeSeriesFilter
from cognite.client.data_classes.filters import Exists
from cognite.client.data_classes.time_series import TimeSeriesProperty
from pydantic import ConfigDict

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client._resource_base import Identifier, RequestResource
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.http_client._item_classes import ItemsResultList
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.dtype_conversion import (
    _EpochConverter,
    _Float64Converter,
    _TextConverter,
    _ValueConverter,
)
from cognite_toolkit._cdf_tk.utils.fileio import SchemaColumn
from cognite_toolkit._cdf_tk.utils.fileio._readers import MultiFileReader
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, TableStorageIO, TableUploadableStorageIO, UploadItem
from .selectors import DataPointsDataSetSelector, DataPointsFileSelector, DataPointsSelector


class DatapointsRequestAdapter(RequestResource):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    datapoints: DataPointInsertionRequest

    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        return {"datapoints": self.datapoints.SerializeToString()}

    def as_id(self) -> Identifier:
        raise NotImplementedError(
            "DatapointsRequestAdapter does not have an identifier. - it wraps multiple timeseries"
        )


class DatapointsIO(
    TableStorageIO[DataPointsSelector, DataPointListResponse],
    TableUploadableStorageIO[DataPointsSelector, DataPointListResponse, DatapointsRequestAdapter],
):
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".csv"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    CHUNK_SIZE = 10_000
    DOWNLOAD_CHUNK_SIZE = 100
    BASE_SELECTOR = DataPointsSelector
    KIND = "Datapoints"
    SUPPORTED_READ_FORMATS = frozenset({".csv"})
    UPLOAD_ENDPOINT = "/timeseries/data"
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = None
    MAX_TOTAL_DATAPOINTS = 10_000_000
    MAX_PER_REQUEST_DATAPOINTS = 100_000
    MAX_PER_REQUEST_DATAPOINTS_AGGREGATION = 10_000

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._warned_columns: set[str] = set()
        self._epoc_converter = _EpochConverter(nullable=True)
        self._numeric_converter = _Float64Converter(nullable=True)
        self._string_converter = _TextConverter(nullable=True)

    def as_id(self, item: DataPointListResponse) -> str:
        raise NotImplementedError()

    def get_schema(self, selector: DataPointsSelector) -> list[SchemaColumn]:
        return [
            SchemaColumn(name="externalId", type="string"),
            SchemaColumn(name="timestamp", type="epoch"),
            SchemaColumn(
                name="value",
                type="string"
                if isinstance(selector, DataPointsDataSetSelector) and selector.data_type == "string"
                else "float",
            ),
        ]

    def stream_data(
        self, selector: DataPointsSelector, limit: int | None = None
    ) -> Iterable[Page[DataPointListResponse]]:
        if not isinstance(selector, DataPointsDataSetSelector):
            raise RuntimeError(
                f"{type(self).__name__} only supports streaming data for DataPointsDataSetSelector selectors. Got {type(selector).__name__}."
            )
        timeseries_count = self.count(selector)
        if limit is not None:
            timeseries_count = min(timeseries_count or 0, limit)
        limit_per_timeseries = (
            (self.MAX_TOTAL_DATAPOINTS // timeseries_count) if timeseries_count else self.MAX_PER_REQUEST_DATAPOINTS
        )
        limit_per_timeseries = min(limit_per_timeseries, self.MAX_PER_REQUEST_DATAPOINTS)
        config = self.client.config
        for timeseries in self.client.time_series(
            data_set_external_ids=[selector.data_set_external_id],
            chunk_size=self.DOWNLOAD_CHUNK_SIZE,
            is_string=True if selector.data_type == "string" else False,
            advanced_filter=Exists(TimeSeriesProperty.external_id),
            limit=limit,
            # We cannot use partitions here as it is not thread safe. This spawn multiple threads
            # that are not shut down until all data is downloaded. We need to be able to abort.
            partitions=None,
        ):
            if not timeseries:
                continue
            # Aggregation of datapoints per timeseries
            items = [
                {
                    "id": ts.id,
                    "start": selector.start,
                    "end": selector.end,
                    "limit": self.MAX_PER_REQUEST_DATAPOINTS_AGGREGATION // len(timeseries),
                    "aggregates": ["count"],
                    "granularity": "1200mo",
                }
                for ts in timeseries
            ]
            response = self.client.http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=config.create_api_url("/timeseries/data/list"),
                    method="POST",
                    accept="application/protobuf",
                    content_type="application/json",
                    body_content={"items": items},  # type: ignore[dict-item]
                )
            )
            if not isinstance(response, SuccessResponse):
                continue
            aggregate_response: DataPointListResponse = DataPointListResponse.FromString(response.content)
            timeseries_ids_with_data: dict[int, int] = {}
            for dp in aggregate_response.items:
                if dp.aggregateDatapoints.datapoints:
                    ts_datapoint_count = int(sum(agg.count for agg in dp.aggregateDatapoints.datapoints))
                    timeseries_ids_with_data[dp.id] = ts_datapoint_count
            total_datapoints = int(sum(timeseries_ids_with_data.values()))
            if total_datapoints == 0:
                continue

            batch: list[dict[str, Any]] = []
            batch_count = 0
            for ts_id, count in timeseries_ids_with_data.items():
                count = min(count, limit_per_timeseries)
                ts_limit = count
                left_over = 0
                if (batch_count + ts_limit) > self.MAX_PER_REQUEST_DATAPOINTS:
                    ts_limit = self.MAX_PER_REQUEST_DATAPOINTS - batch_count
                    left_over = count - ts_limit
                batch.append(
                    {
                        "id": ts_id,
                        "start": selector.start,
                        "end": selector.end,
                        "limit": ts_limit,
                    }
                )
                batch_count += ts_limit
                if batch_count >= self.MAX_PER_REQUEST_DATAPOINTS:
                    if page := self._fetch_datapoints_batch(batch, config):
                        yield page
                    batch = []

                if left_over > 0:
                    batch.append(
                        {
                            "id": ts_id,
                            "start": selector.start,
                            "end": selector.end,
                            "limit": left_over,
                        }
                    )
                    batch_count += left_over
            if batch and (page := self._fetch_datapoints_batch(batch, config)):
                yield page

    def _fetch_datapoints_batch(self, batch: list[dict[str, Any]], config: Any) -> Page[DataPointListResponse] | None:
        response = self.client.http_client.request_single_retries(
            RequestMessage(
                endpoint_url=config.create_api_url("/timeseries/data/list"),
                method="POST",
                accept="application/protobuf",
                content_type="application/json",
                body_content={"items": batch},  # type: ignore[dict-item]
            )
        )
        if not isinstance(response, SuccessResponse):
            return None
        data_response: DataPointListResponse = DataPointListResponse.FromString(response.content)
        return Page("Main", [data_response])

    def count(self, selector: DataPointsSelector) -> int | None:
        if isinstance(selector, DataPointsDataSetSelector):
            return self.client.time_series.aggregate_count(
                filter=TimeSeriesFilter(
                    data_set_ids=[{"externalId": selector.data_set_external_id}],
                    is_string=True if selector.data_type == "string" else False,
                ),
                # We only want time series that have externalID set.
                advanced_filter=Exists(TimeSeriesProperty.external_id),
            )
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[DataPointListResponse], selector: DataPointsSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        raise ToolkitNotImplementedError(
            f"Download of {type(DatapointsIO).__name__.removesuffix('IO')} does not support json format."
        )

    def data_to_row(
        self, data_chunk: Sequence[DataPointListResponse], selector: DataPointsSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        output: list[dict[str, JsonVal]] = []
        for response in data_chunk:
            for item in response.items:
                if item.numericDatapoints.datapoints:
                    for dp in item.numericDatapoints.datapoints:
                        output.append(
                            {
                                "externalId": item.externalId,
                                "timestamp": dp.timestamp,
                                "value": dp.value,
                            }
                        )
                if item.stringDatapoints.datapoints:
                    for dp in item.stringDatapoints.datapoints:
                        output.append(
                            {
                                "externalId": item.externalId,
                                "timestamp": dp.timestamp,
                                "value": dp.value,
                            }
                        )
        return output

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[DatapointsRequestAdapter]],
        http_client: HTTPClient,
        selector: DataPointsSelector | None = None,
    ) -> ItemsResultList:
        results = ItemsResultList()
        for item in data_chunk:
            response = http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=http_client.config.create_api_url(self.UPLOAD_ENDPOINT),
                    method="POST",
                    content_type="application/protobuf",
                    data_content=item.item.datapoints.SerializeToString(),
                )
            )
            results.append(response.as_item_response(item.source_id))
        return results

    def row_to_resource(
        self, source_id: str, row: dict[str, JsonVal], selector: DataPointsSelector | None = None
    ) -> DatapointsRequestAdapter:
        if selector is None:
            raise ValueError("Selector must be provided to convert row to DataPointInsertionItem.")
        # We assume that the row was read using the read_chunks method.
        rows = cast(dict[str, list[Any]], row)
        if isinstance(selector, DataPointsFileSelector):
            datapoints_items = self._rows_to_datapoint_items_file_selector(rows, selector, source_id)
        elif isinstance(selector, DataPointsDataSetSelector):
            datapoints_items = self._rows_to_datapoint_items_data_set_selector(rows, selector, source_id)
        else:
            raise RuntimeError(
                f"Unsupported selector type {type(selector).__name__} for {type(self).__name__}. Trying to transform {source_id!r} from rows to DataPointInsertionRequest."
            )
        return DatapointsRequestAdapter(datapoints=DataPointInsertionRequest(items=datapoints_items))

    def _rows_to_datapoint_items_file_selector(
        self, rows: dict[str, list[Any]], selector: DataPointsFileSelector, source_id: str
    ) -> list[DataPointInsertionItem]:
        if selector.timestamp_column not in rows:
            raise RuntimeError(f"Timestamp column '{selector.timestamp_column}' not found.")

        timestamps = list(
            self._convert_values(
                rows[selector.timestamp_column],
                self._epoc_converter,
                f"timestamps (columns {selector.timestamp_column!r})",
                source_id,
            )
        )

        datapoints_items: list[DataPointInsertionItem] = []
        for col, values in rows.items():
            if col == selector.timestamp_column:
                continue
            column = selector.id_by_column.get(col)
            if column is None:
                self._warn_missing_columns(col)
                continue
            args: dict[str, Any] = column.as_wrapped_id()
            if column.dtype == "numeric":
                number_values = self._convert_values(
                    values, self._numeric_converter, f"numeric datapoints (column {col!r})", source_id
                )
                args["numericDatapoints"] = NumericDatapoints(
                    datapoints=[
                        NumericDatapoint(timestamp=timestamp, value=value)
                        for timestamp, value in zip(timestamps, number_values)
                        if timestamp is not None
                    ]
                )
            elif column.dtype == "string":
                string_values = self._convert_values(
                    values, self._string_converter, f"string datapoints (column {col!r})", source_id
                )
                args["stringDatapoints"] = StringDatapoints(
                    datapoints=[
                        StringDatapoint(timestamp=timestamp, value=value)
                        for timestamp, value in zip(timestamps, string_values)
                        if timestamp is not None
                    ]
                )
            else:
                raise RuntimeError(f"Unsupported dtype {column.dtype} for column {col}.")

            datapoints_items.append(DataPointInsertionItem(**args))

        return datapoints_items

    def _rows_to_datapoint_items_data_set_selector(
        self, rows: dict[str, list[Any]], selector: DataPointsDataSetSelector, source_id: str
    ) -> list[DataPointInsertionItem]:
        if "externalId" not in rows:
            raise RuntimeError("Column 'externalId' not found.")
        if "value" not in rows:
            raise RuntimeError("Column 'value' not found.")
        if "timestamp" not in rows:
            raise RuntimeError("Column 'timestamp' not found.")

        external_ids = rows["externalId"]
        timestamps = list(
            self._convert_values(
                rows["timestamp"],
                self._epoc_converter,
                "timestamps (column 'timestamp')",
                source_id,
            )
        )
        values = list(
            self._convert_values(
                rows["value"],
                self._numeric_converter if selector.data_type == "numeric" else self._string_converter,
                "values (column 'value')",
                source_id,
            )
        )
        sorted_datapoints = sorted(zip(external_ids, timestamps, values), key=lambda x: x[0])
        datapoints_items: list[DataPointInsertionItem] = []
        if selector.data_type == "numeric":
            for external_id, datapoints in groupby(sorted_datapoints, key=lambda x: x[0]):
                datapoints_items.append(
                    DataPointInsertionItem(
                        externalId=external_id,
                        numericDatapoints=NumericDatapoints(
                            datapoints=[
                                NumericDatapoint(timestamp=timestamp, value=value) for _, timestamp, value in datapoints
                            ]
                        ),
                    )
                )
        elif selector.data_type == "string":
            for external_id, datapoints in groupby(sorted_datapoints, key=lambda x: x[0]):
                datapoints_items.append(
                    DataPointInsertionItem(
                        externalId=external_id,
                        stringDatapoints=StringDatapoints(
                            datapoints=[
                                StringDatapoint(timestamp=timestamp, value=value) for _, timestamp, value in datapoints
                            ]
                        ),
                    )
                )
        else:
            raise RuntimeError(f"Unsupported data_type {selector.data_type} for DataPointsDataSetSelector.")

        return datapoints_items

    def _convert_values(
        self, values: list[Any], converter: _ValueConverter, name: str, source_id: str
    ) -> Iterable[Any]:
        failed_count = 0
        for value in values:
            try:
                converted = converter.convert(value)
            except ValueError:
                failed_count += 1
                continue
            yield converted
        if failed_count > 0:
            HighSeverityWarning(
                f"In {source_id}' {failed_count:,} {name} could not be converted and will be skipped."
            ).print_warning(console=self.client.console)

    def _warn_missing_columns(self, column: str) -> None:
        if column not in self._warned_columns:
            HighSeverityWarning(
                f"Column '{column}' not found in selector columns. Skipping this column."
            ).print_warning(console=self.client.console)
            self._warned_columns.add(column)

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> DatapointsRequestAdapter:
        raise ToolkitNotImplementedError(
            f"Upload of {type(DatapointsIO).__name__.removesuffix('IO')} does not support json format."
        )

    @classmethod
    def read_chunks(
        cls, reader: MultiFileReader, selector: DataPointsSelector
    ) -> Iterable[list[tuple[str, dict[str, JsonVal]]]]:
        if not reader.is_table:
            raise RuntimeError(f"{cls.__name__} can only read from TableReader instances.")

        iterator = iter(reader.read_chunks_with_line_numbers())
        try:
            start_row, first = next(iterator)
        except StopIteration:
            # Empty file
            return
        column_names = list(first.keys())
        if isinstance(selector, DataPointsDataSetSelector):
            if set(column_names) != selector.required_columns:
                raise RuntimeError(
                    "When uploading datapoints using a dataset manifest for datapoints, you must have exacatly the "
                    f"columns: {humanize_collection(selector.required_columns)} in the file. Got {humanize_collection(column_names)}. "
                )
        batch: dict[str, list[Any]] = {col: [value] for col, value in first.items()}
        last_row = start_row
        for row_no, chunk in iterator:
            for col, value in chunk.items():
                batch[col].append(value)

            # The number of datapoints is the number of rows times the number of value columns.
            if ((len(column_names) - 1) * len(batch[column_names[0]])) >= cls.CHUNK_SIZE:
                # We cannot guarantee JsonVal here, but that is handled later in the processing pipeline.
                yield [(f"rows {start_row} to {row_no}", batch)]  # type: ignore[list-item]
                start_row = row_no + 1
                batch = {col: [] for col in column_names}
            last_row = row_no
        if any(batch.values()):
            yield [(f"rows {start_row} to{last_row}", batch)]  # type: ignore[list-item]

    @classmethod
    def count_items(cls, reader: MultiFileReader, selector: DataPointsSelector | None = None) -> int:
        if selector is None:
            raise ValueError("A selector is required to correctly count datapoint items.")
        count = reader.count()
        if isinstance(selector, DataPointsFileSelector):
            return count * len(selector.columns)
        return count



import sys
from typing import Any, ClassVar, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    Identifier,
    RequestResource,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata

from .data_modeling import NodeReference
from .identifiers import InternalId, ThreeDModelRevisionId

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class RevisionStatus(BaseModelObject):
    status: Literal["Queued", "Processing", "Done", "Failed"] | None = None
    revision_id: int | None = None
    created_time: int | None = None
    revision_count: int | None = None
    types: list[str] | None = None


class ThreeDModelRequest(RequestResource):
    name: str
    # This field is part of the path request and not the body schema.
    # but is needed for identifier conversion.
    id: int | None = Field(None, exclude=True)

    def as_id(self) -> InternalId:
        if self.id is None:
            raise ValueError("Cannot convert to InternalId when id is None.")
        return InternalId(id=self.id)


class ThreeDModelClassicRequest(ThreeDModelRequest, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata"})
    data_set_id: int | None = None
    metadata: Metadata | None = None


class ThreeDModelDMSRequest(ThreeDModelRequest):
    space: str
    type: Literal["CAD", "PointCloud", "Image360"]
    thumbnail_reference: NodeReference | None = None


class ThreeDModelClassicResponse(ResponseResource[ThreeDModelClassicRequest]):
    name: str
    id: int
    created_time: int
    data_set_id: int | None = None
    metadata: Metadata | None = None
    space: str | None = None
    last_revision_info: RevisionStatus | None = None

    @classmethod
    def request_cls(cls) -> type[ThreeDModelClassicRequest]:
        return ThreeDModelClassicRequest

    def as_id(self) -> InternalId:
        return InternalId(id=self.id)


class ThreeDRevisionCamera(BaseModelObject):
    """Camera settings for a 3D revision (target and position)."""

    target: list[float]
    position: list[float]


class ThreeDRevisionClassicRequest(UpdatableRequestResource):
    """Request class for creating/updating a classic 3D revision."""

    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset(
        {"published", "rotation", "scale", "translation", "camera"}
    )

    published: bool | None = None
    rotation: list[float] | None = None
    scale: list[float] | None = None
    translation: list[float] | None = None
    metadata: Metadata | None = None
    camera: ThreeDRevisionCamera | None = None
    file_id: int
    # This field is used for update/delete and is not part of the create body schema.
    id: int | None = Field(None, exclude=True)
    # model_id is a path parameter, not part of the request body.
    model_id: int = Field(exclude=True)

    def as_id(self) -> ThreeDModelRevisionId:
        if self.id is None:
            raise ValueError("Cannot convert to InternalId when id is None.")
        return ThreeDModelRevisionId(model_id=self.model_id, id=self.id)

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        update = super().as_update(mode)
        # the FileId is immutable.
        if "update" in update and "fileId" in update["update"]:
            del update["update"]["fileId"]
        return update


class ThreeDRevisionClassicResponse(ResponseResource[ThreeDRevisionClassicRequest]):
    """Response class for a classic 3D revision."""

    id: int
    file_id: int
    published: bool
    rotation: list[float] | None = None
    scale: list[float] | None = None
    translation: list[float] | None = None
    camera: ThreeDRevisionCamera | None = None
    status: Literal["Queued", "Processing", "Done", "Failed"]
    metadata: Metadata | None = None
    thumbnail_threed_file_id: int | None = None
    thumbnail_url: str | None = Field(None, alias="thumbnailURL")
    asset_mapping_count: int
    created_time: int
    # model_id is a path parameter, not returned in the API response body.
    model_id: int = Field(-1, exclude=True)

    @classmethod
    def request_cls(cls) -> type[ThreeDRevisionClassicRequest]:
        return ThreeDRevisionClassicRequest

    def as_request_resource(self) -> ThreeDRevisionClassicRequest:
        dumped = self.dump()
        dumped["modelId"] = self.model_id
        return ThreeDRevisionClassicRequest.model_validate(dumped, extra="ignore")


class AssetMappingDMRequest(RequestResource, Identifier):
    node_id: int
    asset_instance_id: NodeReference
    # These fields are part of the path request and not the body schema.
    model_id: int = Field(exclude=True)
    revision_id: int = Field(exclude=True)

    def as_id(self) -> Self:
        return self

    def __str__(self) -> str:
        return f"{self.model_id}_{self.revision_id}_{self.node_id}_{self.asset_instance_id.space}_{self.asset_instance_id.external_id}"


class AssetMappingClassicRequest(RequestResource, Identifier):
    node_id: int
    asset_id: int | None = None
    asset_instance_id: NodeReference | None = None
    # These fields are part of the path request and not the body schema.
    model_id: int = Field(exclude=True)
    revision_id: int = Field(exclude=True)

    def as_id(self) -> Self:
        return self

    def __str__(self) -> str:
        asset_part = (
            f"assetId:{self.asset_id}"
            if self.asset_id is not None
            else f"assetInstance:{self.asset_instance_id.space}_{self.asset_instance_id.external_id}"
            if self.asset_instance_id is not None
            else "noAsset"
        )
        return f"{self.model_id}_{self.revision_id}_{self.node_id}_{asset_part}"


class AssetMappingClassicResponse(ResponseResource[AssetMappingClassicRequest]):
    node_id: int
    asset_id: int | None = None
    asset_instance_id: NodeReference | None = None
    tree_index: int | None = None
    subtree_size: int | None = None
    # These fields are part of the path request and response, but they are included here for convenience.
    model_id: int = Field(-1, exclude=True)
    revision_id: int = Field(-1, exclude=True)

    @classmethod
    def request_cls(cls) -> type[AssetMappingClassicRequest]:
        return AssetMappingClassicRequest

    def as_request_resource(self) -> AssetMappingClassicRequest:
        return AssetMappingClassicRequest.model_validate(
            {**self.dump(), "modelId": self.model_id, "revisionId": self.revision_id}
        )


class AssetMappingDMResponse(ResponseResource[AssetMappingDMRequest]):
    node_id: int
    asset_instance_id: NodeReference
    tree_index: int | None = None
    subtree_size: int | None = None
    # These fields are part of the path request and response, but they are included here for convenience.
    model_id: int = Field(-1, exclude=True)
    revision_id: int = Field(-1, exclude=True)

    @classmethod
    def request_cls(cls) -> type[AssetMappingDMRequest]:
        return AssetMappingDMRequest

    def as_request_resource(self) -> AssetMappingDMRequest:
        return AssetMappingDMRequest.model_validate(
            {**self.dump(), "modelId": self.model_id, "revisionId": self.revision_id}
        )



from collections.abc import Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import SequenceRowFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import SequenceRowId
from cognite_toolkit._cdf_tk.client.resource_classes.sequence_rows import (
    SequenceRowsRequest,
    SequenceRowsResponse,
)


class SequenceRowsAPI(CDFResourceAPI[SequenceRowsResponse]):
    """API for managing sequence row data in CDF.

    This handles inserting, deleting, and retrieving rows from sequences.
    The sequence rows endpoints differ from standard CRUD APIs: insert and delete
    operate on batches of items, while retrieve and retrieve_latest operate on a
    single sequence at a time.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/sequences/data", item_limit=1000, concurrency_max_workers=1),
                "list": Endpoint(method="POST", path="/sequences/data/list", item_limit=1000),
                "delete": Endpoint(
                    method="POST", path="/sequences/data/delete", item_limit=1000, concurrency_max_workers=1
                ),
            },
        )
        self._latest_endpoint = Endpoint(method="POST", path="/sequences/data/latest", item_limit=1000)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[SequenceRowsResponse]:
        return PagedResponse[SequenceRowsResponse].model_validate_json(response.body)

    def _post_single(self, path: str, body: dict[str, Any]) -> str:
        """Send a single-object POST request and return the response body.

        Unlike the items-based endpoints, the retrieve and retrieve_latest
        endpoints accept a single object body (no "items" wrapper).
        """
        request = RequestMessage(
            endpoint_url=self._make_url(path),
            method="POST",
            body_content=body,
        )
        result = self._http_client.request_single_retries(request)
        return result.get_success_or_raise().body

    def create(self, items: Sequence[SequenceRowsRequest]) -> None:
        """Insert rows into one or more sequences.

        This overwrites data in rows and columns that exist.

        Args:
            items: Sequence of SequenceRowsRequest objects, each specifying a
                sequence (by external_id) and the rows/columns to insert.
        """
        self._request_no_response(items, "create")

    def delete(self, items: Sequence[SequenceRowId]) -> None:
        """Delete rows from one or more sequences.

        All columns are affected for the specified row numbers.

        Args:
            items: Sequence of SequenceRowId objects, each specifying a sequence (by external_id) and the row numbers to delete.
        """
        self._request_no_response(items, "delete")

    def latest(self, external_id: str, before: int | None = None) -> SequenceRowsResponse:
        """Retrieve the latest rows for a given sequence.

        This endpoint returns the latest rows for a single sequence, with an optional
        filter for which rows to include. The response structure is the same as the
        list endpoint but without pagination (all rows are returned in one response).

        Args:
            external_id: External ID of the sequence to retrieve rows for.
            before: Optional filter to only include rows with row_number less than this value.

        Returns:
            SequenceRowsResponse containing the latest rows for the specified sequence.
        """
        body: dict[str, Any] = {"externalId": external_id}
        if before is not None:
            body["before"] = before

        result = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._make_url(self._latest_endpoint.path),
                method=self._latest_endpoint.method,
                body_content=body,
                disable_gzip=self._disable_gzip,
                api_version=self._api_version,
            )
        )
        response = result.get_success_or_raise()
        return SequenceRowsResponse.model_validate_json(response.body)

    # Overridden form of the _paginate method to handle the unique pagination structure of sequence rows endpoints
    def _paginate(
        self,
        limit: int,
        cursor: str | None = None,
        params: dict[str, Any] | None = None,
        body: dict[str, Any] | None = None,
        endpoint_path: str | None = None,
    ) -> PagedResponse[SequenceRowsResponse]:
        """Fetch a single page of resources.

        Args:
            params: Query parameters for the request. Supported parameters depend on
                the resource type but typically include:
                - cursor: Cursor for pagination
                - limit: Maximum number of items (defaults to list limit)
                - space: Filter by space
                - includeGlobal: Whether to include global resources
            body : Body content for the request, if applicable.
                limit: Maximum number of items to return in the page.
                cursor: Cursor for pagination.
            limit: Maximum number of items to return in the page.
            cursor: Cursor for pagination.

        Returns:
            A Page containing the items and the cursor for the next page.
        """
        endpoint = self._method_endpoint_map["list"]
        if not (0 < limit <= endpoint.item_limit):
            raise ValueError(f"Limit must be between 1 and {endpoint.item_limit}, got {limit}.")

        body_content: dict[str, Any] = {"limit": limit, **(body or {})}
        if cursor is not None:
            body_content["cursor"] = cursor

        request = RequestMessage(
            endpoint_url=self._make_url(endpoint_path or endpoint.path),
            method=endpoint.method,
            body_content=body_content,
            disable_gzip=self._disable_gzip,
            api_version=self._api_version,
        )
        result = self._http_client.request_single_retries(request)
        response = result.get_success_or_raise()

        parsed = SequenceRowsResponse.model_validate_json(response.body)
        return PagedResponse(items=[parsed], nextCursor=parsed.next_cursor)

    def paginate(
        self, filter: SequenceRowFilter, limit: int = 100, cursor: str | None = None
    ) -> PagedResponse[SequenceRowsResponse]:
        """Iterate over sequence rows in CDF with pagination.

        Args:
            filter: Which sequence rows to paginate over.
            limit: Maximum number of rows to return per page.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of SequenceResponse objects.
        """
        return self._paginate(cursor=cursor, limit=limit, body=filter.dump())

    # Overridden to count on the number of rows rather than number of items.
    def _iterate(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        params: dict[str, Any] | None = None,
        body: dict[str, Any] | None = None,
        endpoint_path: str | None = None,
    ) -> Iterable[list[SequenceRowsResponse]]:
        """Iterate over all resources, handling pagination automatically."""
        next_cursor = cursor
        total = 0
        endpoint = self._method_endpoint_map["list"]
        while True:
            page_limit = endpoint.item_limit if limit is None else min(limit - total, endpoint.item_limit)
            page = self._paginate(
                limit=page_limit, cursor=next_cursor, params=params, body=body, endpoint_path=endpoint_path
            )
            yield page.items
            total += len(page.items[0].rows)
            if page.next_cursor is None or (limit is not None and total >= limit):
                break
            next_cursor = page.next_cursor

    def iterate(self, filter: SequenceRowFilter, limit: int | None = 100) -> Iterable[list[SequenceRowsResponse]]:
        """Iterate over sequence rows in CDF.

        Args:
            filter: Which sequence rows to iterate over.
            limit: Maximum number of rows to return. If None, iterates over all rows (may be slow for large sequences).

        Returns:
            Iterable of lists of SequenceResponse objects.
        """
        return self._iterate(limit=limit, body=filter.dump())

    def list(self, filter: SequenceRowFilter, limit: int | None = 100) -> list[SequenceRowsResponse]:
        """List sequence rows for a given sequence in CDF.

        Args:
            filter: Which sequence rows to list.
            limit: Maximum number of rows to return. If None, returns all rows (may be slow for large sequences).

        Returns:
            List of SequenceResponse objects.
        """
        # The sequence rows list endpoint returns one sequence's rows at a time (a single SequenceRowsResponse
        # per page), so we aggregate all rows into the first response rather than returning multiple responses.
        # Since each page contains exactly one item, we extend first_response.rows with subsequent pages' rows.
        first_response: SequenceRowsResponse | None = None
        for batch in self._iterate(limit=limit, body=filter.dump()):
            if not batch:
                break
            if first_response is None:
                first_response = batch[0]
            else:
                first_response.rows.extend(batch[0].rows)
        return [first_response] if first_response is not None else []



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import TransformationNotificationFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId
from cognite_toolkit._cdf_tk.client.resource_classes.transformation_notification import (
    TransformationNotificationRequest,
    TransformationNotificationResponse,
)


class TransformationNotificationsAPI(CDFResourceAPI[TransformationNotificationResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/transformations/notifications", item_limit=1000),
                "delete": Endpoint(method="POST", path="/transformations/notifications/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/transformations/notifications", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[TransformationNotificationResponse]:
        return PagedResponse[TransformationNotificationResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalId]:
        return ResponseItems[InternalId].model_validate_json(response.body)

    def create(self, items: Sequence[TransformationNotificationRequest]) -> list[TransformationNotificationResponse]:
        """Subscribe for notifications on transformation errors.

        Args:
            items: List of TransformationNotificationRequest objects to create.
        Returns:
            List of created TransformationNotificationResponse objects.
        """
        return self._request_item_response(items, "create")

    def delete(self, items: Sequence[InternalId]) -> None:
        """Delete notification subscriptions by notification ID.

        Args:
            items: List of InternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: TransformationNotificationFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[TransformationNotificationResponse]:
        """Fetch a page of transformation notification subscriptions from CDF.

        Args:
            filter: TransformationNotificationFilter object to filter the results.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of TransformationNotificationResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def iterate(
        self,
        filter: TransformationNotificationFilter | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[TransformationNotificationResponse]]:
        """Iterate over all transformation notification subscriptions in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of TransformationNotificationResponse objects.
        """
        return self._iterate(limit=limit, params=filter.dump() if filter else None)

    def list(
        self,
        filter: TransformationNotificationFilter | None = None,
        limit: int | None = 100,
    ) -> list[TransformationNotificationResponse]:
        """List all transformation notification subscriptions in CDF.

        Returns:
            List of TransformationNotificationResponse objects.
        """
        return self._list(limit=limit, params=filter.dump() if filter else None)



from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.robotics._data_post_processing import (
    RobotDataPostProcessingRequest,
    RobotDataPostProcessingResponse,
)


class DataPostProcessingAPI(CDFResourceAPI[RobotDataPostProcessingResponse]):
    """API for managing DataPostProcessing resources in CDF."""

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(
                    method="POST", path="/robotics/data_postprocessing", item_limit=1000, concurrency_max_workers=1
                ),
                "retrieve": Endpoint(
                    method="POST",
                    path="/robotics/data_postprocessing/byids",
                    item_limit=1000,
                    concurrency_max_workers=1,
                ),
                "update": Endpoint(
                    method="POST",
                    path="/robotics/data_postprocessing/update",
                    item_limit=1000,
                    concurrency_max_workers=1,
                ),
                "delete": Endpoint(
                    method="POST",
                    path="/robotics/data_postprocessing/delete",
                    item_limit=1000,
                    concurrency_max_workers=1,
                ),
                "list": Endpoint(method="GET", path="/robotics/data_postprocessing", item_limit=1000),
            },
            disable_gzip=True,
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[RobotDataPostProcessingResponse]:
        return PagedResponse[RobotDataPostProcessingResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[ExternalId]:
        return ResponseItems[ExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[RobotDataPostProcessingRequest]) -> list[RobotDataPostProcessingResponse]:
        """Create data post-processing configurations in CDF.

        Args:
            items: List of RobotDataPostProcessingRequest objects to create.
        Returns:
            List of created RobotDataPostProcessingResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False
    ) -> list[RobotDataPostProcessingResponse]:
        """Retrieve data post-processing configurations from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: If True, ignores unknown IDs during retrieval.
        Returns:
            List of retrieved RobotDataPostProcessingResponse objects.
        """
        if ignore_unknown_ids:
            return self._request_item_split_retries(items, method="retrieve")
        return self._request_item_response(items, method="retrieve")

    def update(
        self, items: Sequence[RobotDataPostProcessingRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[RobotDataPostProcessingResponse]:
        """Update data post-processing configurations in CDF.

        Args:
            items: List of RobotDataPostProcessingRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated RobotDataPostProcessingResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete data post-processing configurations from CDF.

        Args:
            items: List of ExternalId objects to delete.
            ignore_unknown_ids: If True, ignores unknown IDs during deletion.
        """
        if ignore_unknown_ids:
            self._request_item_split_retries_no_response(items, "delete")
        else:
            self._request_no_response(items, "delete")

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[RobotDataPostProcessingResponse]:
        """Paginate over data post-processing configurations in CDF.

        Args:
            limit: Maximum number of items per page.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of RobotDataPostProcessingResponse objects.
        """
        return self._paginate(cursor=cursor, limit=limit)

    def iterate(
        self,
        limit: int | None = 100,
    ) -> Iterable[list[RobotDataPostProcessingResponse]]:
        """Iterate over all data post-processing configurations in CDF.

        Args:
            limit: Maximum number of items to return. None returns all.

        Returns:
            Iterable of lists of RobotDataPostProcessingResponse objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = 100) -> list[RobotDataPostProcessingResponse]:
        """List all data post-processing configurations in CDF.

        Args:
            limit: Maximum number of items to return. None returns all.

        Returns:
            List of RobotDataPostProcessingResponse objects.
        """
        return self._list(limit=limit)



from collections.abc import Iterable, Sequence
from itertools import zip_longest
from typing import Any

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    HTTPResult,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.http_client._item_classes import ItemsResultList
from cognite_toolkit._cdf_tk.client.resource_classes.chart import ChartRequest, ChartResponse
from cognite_toolkit._cdf_tk.client.resource_classes.legacy.canvas import (
    IndustrialCanvas,
    IndustrialCanvasApply,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, UploadableStorageIO, UploadItem
from .selectors import (
    AllChartsSelector,
    CanvasExternalIdSelector,
    CanvasSelector,
    ChartExternalIdSelector,
    ChartOwnerSelector,
    ChartSelector,
)


class ChartIO(UploadableStorageIO[ChartSelector, ChartResponse, ChartRequest]):
    KIND = "Charts"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = ChartSelector
    UPLOAD_ENDPOINT_TYPE = "app"
    UPLOAD_ENDPOINT_METHOD = "PUT"
    UPLOAD_ENDPOINT = "/storage/charts/charts"

    def as_id(self, item: ChartResponse) -> str:
        return item.external_id

    def stream_data(self, selector: ChartSelector, limit: int | None = None) -> Iterable[Page]:
        selected_charts = self.client.charts.list(visibility="PUBLIC")
        if isinstance(selector, AllChartsSelector):
            ...
        elif isinstance(selector, ChartOwnerSelector):
            selected_charts = [chart for chart in selected_charts if chart.owner_id == selector.owner_id]
        elif isinstance(selector, ChartExternalIdSelector):
            external_id_set = set(selector.external_ids)
            selected_charts = [chart for chart in selected_charts if chart.external_id in external_id_set]
        else:
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for ChartIO")

        if limit is not None:
            selected_charts = selected_charts[:limit]
        for chunk in chunker_sequence(selected_charts, self.CHUNK_SIZE):
            yield Page(worker_id="main", items=chunk)

    def count(self, selector: ChartSelector) -> int | None:
        # There is no way to get the count of charts up front.
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[ChartResponse], selector: ChartSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        self._populate_timeseries_id_cache(data_chunk)
        return [self._dump_resource(chart) for chart in data_chunk]

    def _populate_timeseries_id_cache(self, data_chunk: Sequence[ChartResponse]) -> None:
        timeseries_ids: set[int] = set()
        for chart in data_chunk:
            for item in chart.data.time_series_collection or []:
                if item.ts_id is not None and item.ts_external_id is None:
                    # We only look-up the internalID if the externalId is missing
                    timeseries_ids.add(item.ts_id)
        if timeseries_ids:
            self.client.lookup.time_series.external_id(list(timeseries_ids))

    def _dump_resource(self, chart: ChartResponse) -> dict[str, JsonVal]:
        dumped = chart.as_request_resource().dump()
        if isinstance(data := dumped.get("data"), dict) and isinstance(
            collection := data.get("timeSeriesCollection"), list
        ):
            for item in collection:
                ts_id = item.pop("tsId", None)
                if ts_id and item.get("tsExternalId") is None:
                    # We only look-up the externalID if it is missing
                    ts_external_id = self.client.lookup.time_series.external_id(ts_id)
                    if ts_external_id is not None:
                        item["tsExternalId"] = ts_external_id
        return dumped

    def json_chunk_to_data(
        self, data_chunk: list[tuple[str, dict[str, JsonVal]]]
    ) -> Sequence[UploadItem[ChartRequest]]:
        self._populate_timeseries_external_id_cache([item_json for _, item_json in data_chunk])
        return super().json_chunk_to_data(data_chunk)

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> ChartRequest:
        return self._load_resource(item_json)

    def _populate_timeseries_external_id_cache(self, item_jsons: Sequence[dict[str, JsonVal]]) -> None:
        timeseries_external_ids: set[str] = set()
        for item_json in item_jsons:
            if isinstance(data := item_json.get("data"), dict) and isinstance(
                collection := data.get("timeSeriesCollection"), list
            ):
                for item in collection:
                    if not isinstance(item, dict):
                        continue
                    ts_external_id = item.get("tsExternalId")
                    if isinstance(ts_external_id, str):
                        timeseries_external_ids.add(ts_external_id)
        if timeseries_external_ids:
            self.client.lookup.time_series.id(list(timeseries_external_ids))

    def _load_resource(self, item_json: dict[str, JsonVal]) -> ChartRequest:
        if isinstance(data := item_json.get("data"), dict) and isinstance(
            collection := data.get("timeSeriesCollection"), list
        ):
            for item in collection:
                if not isinstance(item, dict):
                    continue
                ts_external_id = item.get("tsExternalId")
                if isinstance(ts_external_id, str) and item.get("tsId") is None:
                    # We only look-up the internalID if it is missing
                    ts_id = self.client.lookup.time_series.id(ts_external_id)
                    if ts_id is not None:
                        item["tsId"] = ts_id
        return ChartRequest._load(item_json)


class CanvasIO(UploadableStorageIO[CanvasSelector, IndustrialCanvas, IndustrialCanvasApply]):
    """Download and upload Industrial Canvases to/from CDF.

    Args:
        client (ToolkitClient): The Cognite Toolkit client to use for API interactions.
        exclude_existing_version (bool): Whether to exclude the 'existingVersion' field when uploading canvases.
            Defaults to True. If you set this to False, the upload may fail if the existing version in CDF is
            lower or equal to the one in the uploaded data.
    """

    KIND = "IndustrialCanvas"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = CanvasSelector

    def __init__(self, client: ToolkitClient, exclude_existing_version: bool = True) -> None:
        super().__init__(client)
        self.exclude_existing_version = exclude_existing_version

    def as_id(self, item: IndustrialCanvas) -> str:
        return item.as_id()

    def stream_data(self, selector: CanvasSelector, limit: int | None = None) -> Iterable[Page]:
        if not isinstance(selector, CanvasExternalIdSelector):
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for CanvasIO")
        canvas_ids = selector.external_ids
        if limit is not None and len(canvas_ids) > limit:
            canvas_ids = canvas_ids[:limit]

        for chunk in chunker_sequence(canvas_ids, self.CHUNK_SIZE):
            items: list[IndustrialCanvas] = []
            for canvas_id in chunk:
                canvas = self.client.canvas.industrial.retrieve(canvas_id)
                if canvas is not None:
                    items.append(canvas)
                else:
                    MediumSeverityWarning("Canvas with external ID {canvas_id!r} not found. Skipping.").print_warning(
                        console=self.client.console
                    )
            yield Page(worker_id="main", items=items)

    def count(self, selector: CanvasSelector) -> int | None:
        if not isinstance(selector, CanvasExternalIdSelector):
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for CanvasIO")
        return len(selector.external_ids)

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[IndustrialCanvasApply]],
        http_client: HTTPClient,
        selector: CanvasSelector | None = None,
    ) -> ItemsResultList:
        config = http_client.config
        results = ItemsResultList()
        for item in data_chunk:
            instances = item.item.as_instances()
            upsert_items: list[dict[str, JsonValue]] = []
            for instance in instances:
                dumped = instance.dump()
                if self.exclude_existing_version:
                    dumped.pop("existingVersion", None)
                upsert_items.append(dumped)

            canvas = item.item.canvas
            existing = self.client.canvas.industrial.retrieve(canvas.external_id)
            if existing is not None:
                existing_instance_ids = existing.as_write().as_instance_ids(include_solution_tags=False)
                delete_set = set(existing_instance_ids) - set(item.item.as_instance_ids())
                to_delete: list[Any] = [
                    {
                        "space": instance_id.space,
                        "externalId": instance_id.external_id,
                        "instanceType": instance_id.instance_type,
                    }
                    for instance_id in delete_set
                ]
            else:
                to_delete = []

            last_response: HTTPResult | None = None
            for upsert_chunk, delete_chunk in zip_longest(
                chunker_sequence(upsert_items, 1000), chunker_sequence(to_delete, 1000), fillvalue=None
            ):
                body_content: dict[str, JsonValue] = {}
                if upsert_chunk:
                    # MyPy fails do understand that list[dict[str, JsonValue]] is a subtype of JsonValue
                    body_content["items"] = upsert_chunk  # type: ignore[assignment]
                if delete_chunk:
                    body_content["delete"] = delete_chunk

                response = http_client.request_single_retries(
                    message=RequestMessage(
                        endpoint_url=config.create_api_url("/models/instances"),
                        method="POST",
                        body_content=body_content,
                    )
                )
                if not isinstance(response, SuccessResponse):
                    results.append(response.as_item_response(item.source_id))
                last_response = response
            if last_response is not None:
                results.append(last_response.as_item_response(item.source_id))
        return results

    def data_to_json_chunk(
        self, data_chunk: Sequence[IndustrialCanvas], selector: CanvasSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        self._populate_id_cache(data_chunk)
        return [self._dump_resource(canvas) for canvas in data_chunk]

    def _populate_id_cache(self, data_chunk: Sequence[IndustrialCanvas]) -> None:
        """Populate the client's lookup cache with all referenced resources in the canvases."""
        asset_ids: set[int] = set()
        time_series_ids: set[int] = set()
        event_ids: set[int] = set()
        file_ids: set[int] = set()
        for canvas in data_chunk:
            for container_ref in canvas.container_references:
                if container_ref.container_reference_type == "asset":
                    asset_ids.add(container_ref.resource_id)
                elif container_ref.container_reference_type == "timeseries":
                    time_series_ids.add(container_ref.resource_id)
                elif container_ref.container_reference_type == "event":
                    event_ids.add(container_ref.resource_id)
                elif container_ref.container_reference_type == "file":
                    file_ids.add(container_ref.resource_id)
        if asset_ids:
            self.client.lookup.assets.external_id(list(asset_ids))
        if time_series_ids:
            self.client.lookup.time_series.external_id(list(time_series_ids))
        if event_ids:
            self.client.lookup.events.external_id(list(event_ids))
        if file_ids:
            self.client.lookup.files.external_id(list(file_ids))

    def _dump_resource(self, canvas: IndustrialCanvas) -> dict[str, JsonVal]:
        dumped = canvas.as_write().dump(keep_existing_version=False)
        references = dumped.get("containerReferences", [])
        if not isinstance(references, list):
            return dumped
        new_container_references: list[Any] = []
        for container_ref in references:
            if not isinstance(container_ref, dict):
                new_container_references.append(container_ref)
                continue
            sources = container_ref.get("sources", [])
            if not isinstance(sources, list) or len(sources) == 0:
                new_container_references.append(container_ref)
                continue
            source = sources[0]
            if not isinstance(source, dict) or "properties" not in source:
                new_container_references.append(container_ref)
                continue
            properties = source["properties"]
            if not isinstance(properties, dict):
                new_container_references.append(container_ref)
                continue
            reference_type = properties.get("containerReferenceType")
            if (
                reference_type
                in {
                    "charts",
                    "dataGrid",
                }
            ):  # These container reference types are special cases with a resourceId statically set to -1, which is why we skip them
                new_container_references.append(container_ref)
                continue
            resource_id = properties.pop("resourceId", None)
            if not isinstance(resource_id, int):
                HighSeverityWarning(
                    f"Invalid resourceId {resource_id!r} in Canvas {canvas.canvas.name}. Skipping."
                ).print_warning(console=self.client.console)
                continue
            if reference_type == "asset":
                external_id = self.client.lookup.assets.external_id(resource_id)
            elif reference_type == "timeseries":
                external_id = self.client.lookup.time_series.external_id(resource_id)
            elif reference_type == "event":
                external_id = self.client.lookup.events.external_id(resource_id)
            elif reference_type == "file":
                external_id = self.client.lookup.files.external_id(resource_id)
            else:
                new_container_references.append(container_ref)
                continue
            if external_id is None:
                HighSeverityWarning(
                    f"Failed to look-up {reference_type} external ID for resource ID {resource_id!r}. Skipping resource in Canvas {canvas.canvas.name}"
                ).print_warning(console=self.client.console)
                continue
            properties["resourceExternalId"] = external_id
            new_container_references.append(container_ref)
        dumped["containerReferences"] = new_container_references
        return dumped

    def json_chunk_to_data(
        self, data_chunk: list[tuple[str, dict[str, JsonVal]]]
    ) -> Sequence[UploadItem[IndustrialCanvasApply]]:
        self._populate_external_id_cache([item_json for _, item_json in data_chunk])
        return super().json_chunk_to_data(data_chunk)

    def _populate_external_id_cache(self, item_jsons: Sequence[dict[str, JsonVal]]) -> None:
        """Populate the client's lookup cache with all referenced resources in the canvases."""
        asset_external_ids: set[str] = set()
        time_series_external_ids: set[str] = set()
        event_external_ids: set[str] = set()
        file_external_ids: set[str] = set()
        for item_json in item_jsons:
            references = item_json.get("containerReferences", [])
            if not isinstance(references, list):
                continue
            for container_ref in references:
                if not isinstance(container_ref, dict):
                    continue
                sources = container_ref.get("sources", [])
                if not isinstance(sources, list) or len(sources) == 0:
                    continue
                source = sources[0]
                if not isinstance(source, dict) or "properties" not in source:
                    continue
                properties = source["properties"]
                if not isinstance(properties, dict):
                    continue

                resource_external_id = properties.get("resourceExternalId")
                if not isinstance(resource_external_id, str):
                    continue

                reference_type = properties.get("containerReferenceType")
                if reference_type == "asset":
                    asset_external_ids.add(resource_external_id)
                elif reference_type == "timeseries":
                    time_series_external_ids.add(resource_external_id)
                elif reference_type == "event":
                    event_external_ids.add(resource_external_id)
                elif reference_type == "file":
                    file_external_ids.add(resource_external_id)

        if asset_external_ids:
            self.client.lookup.assets.id(list(asset_external_ids))
        if time_series_external_ids:
            self.client.lookup.time_series.id(list(time_series_external_ids))
        if event_external_ids:
            self.client.lookup.events.id(list(event_external_ids))
        if file_external_ids:
            self.client.lookup.files.id(list(file_external_ids))

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> IndustrialCanvasApply:
        return self._load_resource(item_json)

    def _load_resource(self, item_json: dict[str, JsonVal]) -> IndustrialCanvasApply:
        name = self._get_name(item_json)
        references = item_json.get("containerReferences", [])
        if not isinstance(references, list):
            return IndustrialCanvasApply._load(item_json)
        new_container_references: list[Any] = []
        for container_ref in references:
            if not isinstance(container_ref, dict):
                new_container_references.append(container_ref)
                continue
            sources = container_ref.get("sources", [])
            if not isinstance(sources, list) or len(sources) == 0:
                new_container_references.append(container_ref)
                continue
            source = sources[0]
            if not isinstance(source, dict) or "properties" not in source:
                new_container_references.append(container_ref)
                continue
            properties = source["properties"]
            if not isinstance(properties, dict):
                new_container_references.append(container_ref)
                continue
            resource_external_id = properties.pop("resourceExternalId", None)
            if not isinstance(resource_external_id, str):
                new_container_references.append(container_ref)
                continue
            reference_type = properties.get("containerReferenceType")
            if reference_type == "asset":
                resource_id = self.client.lookup.assets.id(resource_external_id)
            elif reference_type == "timeseries":
                resource_id = self.client.lookup.time_series.id(resource_external_id)
            elif reference_type == "event":
                resource_id = self.client.lookup.events.id(resource_external_id)
            elif reference_type == "file":
                resource_id = self.client.lookup.files.id(resource_external_id)
            else:
                new_container_references.append(container_ref)
                continue
            if resource_id is None:
                # Failed look-up, skip the resourceId setting
                HighSeverityWarning(
                    f"Failed to look-up {reference_type} ID for external ID {resource_external_id!r}. Skipping resource in Canvas {name}"
                ).print_warning(console=self.client.console)
                continue
            properties["resourceId"] = resource_id
            new_container_references.append(container_ref)
        new_item = dict(item_json)
        new_item["containerReferences"] = new_container_references

        return IndustrialCanvasApply._load(new_item)

    @classmethod
    def _get_name(cls, item_json: dict[str, JsonVal]) -> str:
        try:
            return item_json["canvas"]["sources"][0]["properties"]["name"]  # type: ignore[index,return-value, call-overload]
        except (KeyError, IndexError, TypeError):
            return "<unknown>"



from collections.abc import Iterable, Sequence
from typing import Any, Literal

from cognite_toolkit._cdf_tk.client.api.transformation_notifications import TransformationNotificationsAPI
from cognite_toolkit._cdf_tk.client.api.transformation_schedules import TransformationSchedulesAPI
from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import TransformationFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.transformation import (
    SQLQueryResponse,
    TransformationRequest,
    TransformationResponse,
)


class TransformationsAPI(CDFResourceAPI[TransformationResponse]):
    DEFAULT_TIMEOUT_RUN_QUERY = 240.0  # seconds, this is the maximum timeout for running queries in CDF

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/transformations", item_limit=1000),
                "retrieve": Endpoint(method="POST", path="/transformations/byids", item_limit=1000),
                "update": Endpoint(method="POST", path="/transformations/update", item_limit=1000),
                "delete": Endpoint(method="POST", path="/transformations/delete", item_limit=1000),
                "list": Endpoint(method="POST", path="/transformations/filter", item_limit=1000),
            },
        )
        self.schedules = TransformationSchedulesAPI(http_client)
        self.notifications = TransformationNotificationsAPI(http_client)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[TransformationResponse]:
        return PagedResponse[TransformationResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[TransformationRequest]) -> list[TransformationResponse]:
        """Create transformations in CDF.

        Args:
            items: List of TransformationRequest objects to create.
        Returns:
            List of created TransformationResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False, with_job_details: bool = False
    ) -> list[TransformationResponse]:
        """Retrieve transformations from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
            with_job_details: Whether the transformations will be returned with running job and last created job details.
        Returns:
            List of retrieved TransformationResponse objects.
        """
        return self._request_item_response(
            items,
            method="retrieve",
            extra_body={"ignoreUnknownIds": ignore_unknown_ids, "withJobDetails": with_job_details},
        )

    def update(
        self, items: Sequence[TransformationRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[TransformationResponse]:
        """Update transformations in CDF.

        Args:
            items: List of TransformationRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated TransformationResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete transformations from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def preview(
        self,
        query: str | None = None,
        convert_to_string: bool = False,
        limit: int | None = 100,
        source_limit: int | None = 100,
        infer_schema_limit: int | None = 10_000,
        timeout: float | None = DEFAULT_TIMEOUT_RUN_QUERY,
    ) -> SQLQueryResponse:
        """`Preview the result of a query. <https://developer.cognite.com/api#tag/Query/operation/runPreview>`_

        Toolkit runs long-running queries that takes longer than the typical default of 30 seconds. In addition,
        we do not want to retry, which typically up to 10 times, as the user will have to wait for a long time. Instead,
        we want to fail provide the user with the error and then let the user decide whether to retry or not by
        running the CLI command again.

        Args:
            query (str | None): SQL query to run for preview.
            convert_to_string (bool): Stringify values in the query results, default is False.
            limit (int | None): Maximum number of rows to return in the final result, default is 100.
            source_limit (int | None): Maximum number of items to read from the data source or None to run without limit, default is 100.
            infer_schema_limit (int | None): Limit for how many rows that are used for inferring result schema, default is 10 000.
            timeout (int | None): Number of seconds to wait before cancelling a query. The default, and maximum, is 240.

        Returns:
            SQLQueryResponse: Result of the executed query
        """
        body: dict[str, Any] = {
            "query": query,
            "convertToString": convert_to_string,
        }
        if limit is not None:
            body["limit"] = limit
        if source_limit is not None:
            body["sourceLimit"] = source_limit
        if infer_schema_limit is not None:
            body["inferSchemaLimit"] = infer_schema_limit
        if timeout is not None:
            # This is the server-side timeout for how long the query is allowed to run before it is cancelled.
            body["timeout"] = timeout

        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._http_client.config.create_api_url("/transformations/query/run"),
                method="POST",
                body_content=body,
                client_timeout=timeout
                if timeout is not None
                else (self.DEFAULT_TIMEOUT_RUN_QUERY + 60),  # add a buffer to the timeout
                retry_status=False,
            )
        ).get_success_or_raise()
        return SQLQueryResponse.model_validate_json(response.body)

    def paginate(
        self,
        filter: TransformationFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[TransformationResponse]:
        """Iterate over all transformations in CDF.

        Args:
            filter: TransformationFilter object to filter transformations.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of TransformationResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": filter.model_dump(by_alias=True) if filter else {}},
        )

    def iterate(
        self,
        filter: TransformationFilter | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[TransformationResponse]]:
        """Iterate over all transformations in CDF.

        Args:
            filter: TransformationFilter object to filter transformations.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of TransformationResponse objects.
        """
        return self._iterate(
            limit=limit,
            body={"filter": filter.model_dump(by_alias=True) if filter else {}},
        )

    def list(
        self,
        filter: TransformationFilter | None = None,
        limit: int | None = 100,
    ) -> list[TransformationResponse]:
        """List all transformations in CDF.

        Args:
            filter: TransformationFilter object to filter transformations.
            limit: Maximum number of items to return.

        Returns:
            List of TransformationResponse objects.
        """
        return self._list(limit=limit, body={"filter": filter.model_dump(by_alias=True) if filter else {}})



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from .base import ToolkitResource


class SimulatorModelRevisionYAML(ToolkitResource):
    """Simulator model revision YAML resource class.

    Based on: https://api-docs.cognite.com/20230101/tag/Simulator-Models/operation/create_simulator_model_revision_simulators_models_revisions_post
    """

    external_id: str = Field(description="External ID of the simulator model revision.", min_length=1, max_length=255)
    model_external_id: str = Field(description="External ID of the simulator model.")
    description: str | None = Field(None, description="Description of the simulator model revision.", max_length=255)
    file_external_id: str = Field(description="External ID of the file containing the model.")

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



from pathlib import Path
from typing import Annotated, Any

import questionary
import typer
from questionary import Choice

from cognite_toolkit._cdf_tk.commands import UploadCommand
from cognite_toolkit._cdf_tk.constants import DATA_DEFAULT_DIR, DATA_MANIFEST_SUFFIX, DATA_RESOURCE_DIR
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

DEFAULT_INPUT_DIR = Path.cwd() / DATA_DEFAULT_DIR


class UploadApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.upload_main)
        self.command("dir")(self.upload_dir)

    @staticmethod
    def upload_main(ctx: typer.Context) -> None:
        """Commands to upload data to CDF."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf upload --help[/] for more information.")
        return None

    @staticmethod
    def upload_dir(
        ctx: typer.Context,
        input_dir: Annotated[
            Path | None,
            typer.Argument(
                help="The directory containing the data to upload. If not specified, an interactive prompt will ask for the directory.",
                exists=True,
                file_okay=False,
                dir_okay=True,
                resolve_path=True,
            ),
        ] = None,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the command will not actually upload the data, but will print what would be uploaded.",
            ),
        ] = False,
        deploy_resources: Annotated[
            bool,
            typer.Option(
                "--deploy-resources",
                "-r",
                help="If set, the command will look for resource configuration files in adjacent folders and create them if they do not exist.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Commands to upload data to CDF."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = UploadCommand(client=client)
        if input_dir is None:
            input_candidate = sorted({p.parent for p in DEFAULT_INPUT_DIR.rglob(f"**/*{DATA_MANIFEST_SUFFIX}")})
            if not input_candidate:
                typer.echo(f"No data manifests found in default directory: {DEFAULT_INPUT_DIR}")
                raise typer.Exit(code=1)
            input_dir = questionary.select(
                "Select the input directory containing the data to upload:",
                choices=[Choice(str(option.name), value=option) for option in input_candidate],
            ).unsafe_ask()
            dry_run = questionary.confirm("Proceed with dry run?", default=dry_run).unsafe_ask()
            if dry_run is None:
                typer.echo("No selection made for dry run. Exiting.")
                raise typer.Exit(code=1)
            resource_dir = Path(input_dir) / DATA_RESOURCE_DIR
            if resource_dir.exists():
                if resource_dir.is_relative_to(Path.cwd()):
                    display_name = resource_dir.relative_to(Path.cwd()).as_posix()
                else:
                    display_name = resource_dir.as_posix()

                deploy_resources = questionary.confirm(
                    f"Deploy resources found in {display_name!r}?", default=deploy_resources
                ).unsafe_ask()

        cmd.run(
            lambda: cmd.upload(
                input_dir=input_dir,
                dry_run=dry_run,
                verbose=verbose,
                deploy_resources=deploy_resources,
                client=client,
            )
        )



from collections.abc import Hashable, Sequence
from graphlib import TopologicalSorter
from pathlib import Path
from typing import overload

from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError
from rich import print
from rich.markup import escape
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client._resource_base import T_Identifier, T_RequestResource, T_ResponseResource
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.clean import CleanCommand
from cognite_toolkit._cdf_tk.constants import (
    _RUNNING_IN_BROWSER,
    BUILD_ENVIRONMENT_FILE,
    DATA_UPLOAD_URL,
    HINT_LEAD_TEXT,
)
from cognite_toolkit._cdf_tk.cruds import (
    DataCRUD,
    Loader,
    RawDatabaseCRUD,
    ResourceContainerCRUD,
    ResourceCRUD,
    ResourceWorker,
)
from cognite_toolkit._cdf_tk.cruds._resource_cruds import SimulatorModelRevisionCRUD, SimulatorRoutineRevisionCRUD
from cognite_toolkit._cdf_tk.cruds._worker import CategorizedResources
from cognite_toolkit._cdf_tk.data_classes import (
    BuildEnvironment,
    DatapointDeployResult,
    DeployResult,
    DeployResults,
    ResourceContainerDeployResult,
    ResourceDeployResult,
    UploadDeployResult,
)
from cognite_toolkit._cdf_tk.data_classes._module_directories import ReadModule
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceCreationError,
    ResourceUpdateError,
    ToolkitDeployResourceError,
    ToolkitFileNotFoundError,
    ToolkitNotADirectoryError,
)
from cognite_toolkit._cdf_tk.tk_warnings import EnvironmentVariableMissingWarning
from cognite_toolkit._cdf_tk.tk_warnings.base import WarningList, catch_warnings
from cognite_toolkit._cdf_tk.tk_warnings.other import (
    HighSeverityWarning,
    LowSeverityWarning,
    ToolkitDependenciesIncludedWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection, read_yaml_file
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

from ._utils import _print_ids_or_length


class DeployCommand(ToolkitCommand):
    def __init__(
        self,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
        client: ToolkitClient | None = None,
    ) -> None:
        super().__init__(print_warning, skip_tracking, silent, client)
        self._clean_command = CleanCommand(print_warning, skip_tracking=True)

    def deploy_build_directory(
        self,
        env_vars: EnvironmentVariables,
        build_dir: Path,
        build_env_name: str | None,
        dry_run: bool,
        drop: bool,
        drop_data: bool,
        force_update: bool,
        include: list[str] | None,
        verbose: bool,
    ) -> None:
        include = self._clean_command.validate_include(include)

        build = self._load_build(build_dir, build_env_name)

        client = env_vars.get_client(build.is_strict_validation)

        self._additional_tracking_info.project = client.config.project
        self._additional_tracking_info.cluster = client.config.cdf_cluster

        if not dry_run:
            for module in build.read_modules:
                if module.module_id:
                    self._additional_tracking_info.module_ids.add(module.module_id)
                if module.package_id:
                    self._additional_tracking_info.package_ids.add(module.package_id)

        selected_loaders = self._clean_command.get_selected_loaders(build_dir, build.read_resource_folders, include)
        ordered_loaders = self._order_loaders(selected_loaders, build_dir)
        self._start_message(build_dir, dry_run, env_vars)
        results = DeployResults([], "deploy", dry_run=dry_run)
        if drop or drop_data:
            self.clean_all_resources(
                client, results, build, build_dir, drop, drop_data, dry_run, env_vars, ordered_loaders, verbose
            )

        self.deploy_all_resources(
            client=client,
            build_dir=build_dir,
            env_vars=env_vars,
            results=results,
            dry_run=dry_run,
            has_dropped=drop,
            has_dropped_data=drop_data,
            force_update=force_update,
            ordered_loaders=ordered_loaders,
            build=build,
            verbose=verbose,
        )

        if results.has_counts:
            print(results.counts_table())
        if results.has_uploads:
            print(results.uploads_table())

    def _load_build(self, build_dir: Path, build_env_name: str | None) -> BuildEnvironment:
        if not build_dir.is_dir():
            raise ToolkitNotADirectoryError(
                "The build directory does not exists. Did you forget to run `cdf build` first?"
            )
        build_environment_file_path = build_dir / BUILD_ENVIRONMENT_FILE
        if not build_environment_file_path.is_file():
            raise ToolkitFileNotFoundError(
                f"Could not find build environment file '{BUILD_ENVIRONMENT_FILE}' in '{build_dir}'. "
                "Did you forget to run `cdf build` first?"
            )
        build = BuildEnvironment.load(read_yaml_file(build_environment_file_path), build_env_name, "deploy")
        build.set_environment_variables()
        errors = build.check_source_files_changed()
        for error in errors:
            self.warn(error)
        if errors:
            raise ToolkitDeployResourceError(
                "One or more source files have been modified since the last build. Please rebuild the project."
            )
        return build

    def _order_loaders(
        self, selected_loaders: dict[type[Loader], frozenset[type[Loader]]], build_dir: Path
    ) -> list[type[Loader]]:
        """The loaders must be topological sorted to ensure that dependencies are deployed in the correct order."""
        ordered_loaders: list[type[Loader]] = []
        should_include: list[type[Loader]] = []
        # The topological sort can include loaders that are not selected, so we need to check for that.
        for loader_cls in TopologicalSorter(selected_loaders).static_order():
            if loader_cls in selected_loaders:
                ordered_loaders.append(loader_cls)
            elif (build_dir / loader_cls.folder_name).is_dir():
                should_include.append(loader_cls)
            # Otherwise, it is not in the build directory and not selected, so we skip it.
            # There should be a warning in the build step if it is missing.
        if should_include:
            self.warn(ToolkitDependenciesIncludedWarning(list({item.folder_name for item in should_include})))
        return ordered_loaders

    @staticmethod
    def _start_message(build_dir: Path, dry_run: bool, env_vars: EnvironmentVariables) -> None:
        environment_vars = ""
        if not _RUNNING_IN_BROWSER:
            environment_vars = f"\n\nConnected to {env_vars.as_string()}"
        verb = "Checking" if dry_run else "Deploying"
        print(
            Panel(
                f"[bold]{verb}[/]resource files from {build_dir} directory.{environment_vars}",
                expand=False,
            )
        )

    def clean_all_resources(
        self,
        client: ToolkitClient,
        results: DeployResults,
        build: BuildEnvironment,
        build_dir: Path,
        drop: bool,
        drop_data: bool,
        dry_run: bool,
        env_vars: EnvironmentVariables,
        ordered_loaders: list[type[Loader]],
        verbose: bool,
    ) -> None:
        # Drop has to be done in the reverse order of deploy.
        if drop and drop_data:
            print(Panel("[bold] Cleaning resources as --drop and --drop-data are passed[/]"))
        elif drop:
            print(Panel("[bold] Cleaning resources as --drop is passed[/]"))
        elif drop_data:
            print(Panel("[bold] Cleaning resources as --drop-data is passed[/]"))
        else:
            return None

        for loader_cls in reversed(ordered_loaders):
            if not issubclass(loader_cls, ResourceCRUD):
                continue
            loader: ResourceCRUD = loader_cls.create_loader(client, build_dir)
            if isinstance(loader, SimulatorRoutineRevisionCRUD | SimulatorModelRevisionCRUD):
                # There are deleted when the routine/model is deleted, so skip here.
                continue
            result = self._clean_command.clean_resources(
                loader,
                env_vars=env_vars,
                read_modules=build.read_modules,
                drop=drop,
                dry_run=dry_run,
                drop_data=drop_data,
                verbose=verbose,
            )
            if result:
                results[result.name] = result
        print("[bold]...cleaning complete![/]")
        return None

    @overload
    def deploy_all_resources(
        self,
        client: ToolkitClient,
        build_dir: Path,
        env_vars: EnvironmentVariables,
        results: None = None,
        dry_run: bool = False,
        has_dropped: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        ordered_loaders: list[type[Loader]] | None = None,
        build: BuildEnvironment | None = None,
        verbose: bool = False,
    ) -> DeployResults: ...

    @overload
    def deploy_all_resources(
        self,
        client: ToolkitClient,
        build_dir: Path,
        env_vars: EnvironmentVariables,
        results: DeployResults,
        dry_run: bool = False,
        has_dropped: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        ordered_loaders: list[type[Loader]] | None = None,
        build: BuildEnvironment | None = None,
        verbose: bool = False,
    ) -> None: ...

    def deploy_all_resources(
        self,
        client: ToolkitClient,
        build_dir: Path,
        env_vars: EnvironmentVariables,
        results: DeployResults | None = None,
        dry_run: bool = False,
        has_dropped: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        ordered_loaders: list[type[Loader]] | None = None,
        build: BuildEnvironment | None = None,
        verbose: bool = False,
    ) -> DeployResults | None:
        """Deploy all resources in the build directory.

        Args:
            client: The Cognite client to use for deployment.
            build_dir: The directory containing the built resource files.
            env_vars: The environment variables to use for variable replacement in the resource files.
            results: An optional DeployResults object to store the results of the deployment. If None, a new object will be created and returned.
            dry_run: If True, performs a dry run without actually deploying the resources.
            has_dropped: If True, indicates that resources have been dropped prior to deployment.
            has_dropped_data: If True, indicates that data resources have been dropped prior to deployment.
            force_update: If True, forces an update of resources even if they are unchanged.
            ordered_loaders: An optional list of loader classes to use for deployment. If None, all loaders will be used in the order determined by dependencies.
            build: An optional BuildEnvironment object containing information about the build. If None, data uploads will be skipped.
            verbose: If True, prints detailed information about the deployment process.

        Returns:
            A DeployResults object containing the results of the deployment if results is None, otherwise None.

        """
        if verbose:
            print(Panel("[bold]DEPLOYING resources...[/]"))

        if ordered_loaders is None:
            selected_loaders = self._clean_command.get_selected_loaders(build_dir, set(), None)
            ordered_loaders = self._order_loaders(selected_loaders, build_dir)
        read_modules: list[ReadModule] | None = None
        if build is not None:
            read_modules = build.read_modules
        output_results = DeployResults([], "deploy", dry_run=dry_run) if results is None else results
        for loader_cls in ordered_loaders:
            if issubclass(loader_cls, DataCRUD):
                self.warn(
                    HighSeverityWarning(
                        f"Uploading {loader_cls.kind} data is deprecated and will be removed in v0.8. "
                        f"Use the `cdf data upload dir` command instead. See [{DATA_UPLOAD_URL}]({DATA_UPLOAD_URL}) for more information about "
                        "the upload command."
                    )
                )

            loader = loader_cls.create_loader(client, build_dir)
            resource_result: DeployResult | None
            if isinstance(loader, ResourceCRUD):
                resource_result = self.deploy_resource_type(
                    loader,
                    env_vars,
                    read_modules,
                    dry_run,
                    has_dropped,
                    has_dropped_data,
                    force_update,
                    verbose,
                )
            elif isinstance(loader, DataCRUD) and build is not None:
                resource_result = self.upload_data(loader, build, dry_run, verbose)
            elif isinstance(loader, DataCRUD) and build is None:
                self.warn(
                    LowSeverityWarning(
                        f"Skipping data upload for {loader.display_name} as it is not supported in this context."
                    )
                )
                continue
            else:
                raise ValueError(f"Unsupported loader type {type(loader)}.")

            if resource_result:
                output_results[resource_result.name] = resource_result
            if verbose:
                print("")  # Extra newline
        if results is None:
            return output_results
        return None

    def deploy_resource_type(
        self,
        loader: ResourceCRUD[T_Identifier, T_RequestResource, T_ResponseResource],
        env_vars: EnvironmentVariables,
        read_modules: list[ReadModule] | None = None,
        dry_run: bool = False,
        has_done_drop: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        verbose: bool = False,
    ) -> ResourceDeployResult | None:
        worker = ResourceWorker(loader, "deploy")
        files = worker.load_files(read_modules=read_modules)
        if not files:
            return None

        with catch_warnings(EnvironmentVariableMissingWarning) as env_var_warnings:
            resources = worker.prepare_resources(
                files,
                environment_variables=env_vars.dump(include_os=True),
                is_dry_run=dry_run,
                force_update=force_update,
                verbose=verbose,
            )
        if env_var_warnings:
            print(str(env_var_warnings))
            self.warning_list.extend(env_var_warnings)

        # We are not counting to_delete as these are captured by to_create.
        # (to_delete is used for resources that does not support update and instead needs to be deleted and recreated)
        nr_of_items = len(resources.to_create) + len(resources.to_update) + len(resources.unchanged)
        if nr_of_items == 0:
            return ResourceDeployResult(name=loader.display_name)

        prefix = "Would deploy" if dry_run else "Deploying"
        print(f"[bold]{prefix} {nr_of_items} {loader.display_name} to CDF...[/]")

        # Moved here to avoid printing before the above message.
        if not isinstance(loader, RawDatabaseCRUD):
            for duplicate in worker.duplicates:
                self.warn(LowSeverityWarning(f"Skipping duplicate {loader.display_name} {duplicate}."))

        if dry_run:
            result = self.dry_run_deploy(
                resources,
                loader,
                has_done_drop,
                has_dropped_data,
            )
        else:
            result = self.actual_deploy(resources, loader, env_var_warnings)

        if verbose:
            self._verbose_print(resources, loader, dry_run)

        if isinstance(loader, ResourceContainerCRUD):
            return ResourceContainerDeployResult.from_resource_deploy_result(
                result,
                item_name=loader.item_name,
            )
        return result

    def actual_deploy(
        self,
        resources: CategorizedResources[T_Identifier, T_RequestResource],
        loader: ResourceCRUD[T_Identifier, T_RequestResource, T_ResponseResource],
        env_var_warnings: WarningList | None = None,
    ) -> ResourceDeployResult:
        environment_variable_warning_by_id = {
            identifier: warning
            for warning in env_var_warnings or []
            if isinstance(warning, EnvironmentVariableMissingWarning)
            for identifier in warning.identifiers or []
        }
        nr_of_unchanged = len(resources.unchanged)
        nr_of_deleted, nr_of_created, nr_of_changed = 0, 0, 0
        if resources.to_delete:
            deleted = loader.delete(resources.to_delete)
            nr_of_deleted += deleted
        if resources.to_create:
            created = self._create_resources(resources.to_create, loader, environment_variable_warning_by_id)
            nr_of_created += created
        if resources.to_update:
            updated = self._update_resources(resources.to_update, loader, environment_variable_warning_by_id)
            nr_of_changed += updated
        return ResourceDeployResult(
            name=loader.display_name,
            created=nr_of_created,
            deleted=nr_of_deleted,
            changed=nr_of_changed,
            unchanged=nr_of_unchanged,
            total=nr_of_created + nr_of_deleted + nr_of_changed + nr_of_unchanged,
        )

    @staticmethod
    def dry_run_deploy(
        resources: CategorizedResources[T_Identifier, T_RequestResource],
        loader: ResourceCRUD[T_Identifier, T_RequestResource, T_ResponseResource],
        has_done_drop: bool,
        has_dropped_data: bool,
    ) -> ResourceDeployResult:
        if (
            loader.support_drop
            and has_done_drop
            and (not isinstance(loader, ResourceContainerCRUD) or has_dropped_data)
        ):
            # Means the resources will be deleted and not left unchanged or changed
            for item in resources.unchanged:
                # We cannot use extents as LoadableNodes cannot be extended.
                resources.to_create.append(item)
            for item in resources.to_update:
                resources.to_create.append(item)
            resources.unchanged.clear()
            resources.to_update.clear()
        return ResourceDeployResult(
            name=loader.display_name,
            created=len(resources.to_create),
            deleted=len(resources.to_delete),
            changed=len(resources.to_update),
            unchanged=len(resources.unchanged),
            total=len(resources.to_create)
            + len(resources.to_delete)
            + len(resources.to_update)
            + len(resources.unchanged),
        )

    @staticmethod
    def _verbose_print(
        resources: CategorizedResources[T_Identifier, T_RequestResource],
        loader: ResourceCRUD,
        dry_run: bool,
    ) -> None:
        print_outs = []
        prefix = "Would have " if dry_run else ""
        if resources.to_create:
            print_outs.append(f"{prefix}Created {_print_ids_or_length(loader.get_ids(resources.to_create), limit=20)}")
        if resources.to_update:
            print_outs.append(f"{prefix}Updated {_print_ids_or_length(loader.get_ids(resources.to_update), limit=20)}")
        if resources.unchanged:
            print_outs.append(
                f"{'Untouched' if dry_run else 'Unchanged'} {_print_ids_or_length(loader.get_ids(resources.unchanged), limit=5)}"
            )
        prefix_message = f" {loader.display_name}: "
        if len(print_outs) == 1:
            print(f"{prefix_message}{print_outs[0]}")
        elif len(print_outs) == 2:
            print(f"{prefix_message}{print_outs[0]} and {print_outs[1]}")
        else:
            print(f"{prefix_message}{', '.join(print_outs[:-1])} and {print_outs[-1]}")

    def _create_resources(
        self,
        resources: Sequence[T_RequestResource],
        loader: ResourceCRUD,
        environment_variable_warning_by_id: dict[Hashable, EnvironmentVariableMissingWarning],
    ) -> int:
        try:
            created = loader.create(resources)
        except CogniteAPIError as e:
            message = f"Failed to create resource(s). Error: {escape(str(e))!s}."
            if hint := self._environment_variable_hint(loader.get_ids(resources), environment_variable_warning_by_id):
                message += hint
            raise ResourceCreationError(message) from e
        except CogniteDuplicatedError as e:
            self.warn(
                LowSeverityWarning(
                    f"{len(e.duplicated)} out of {len(resources)} resource(s) already exist(s). {len(e.successful or [])} resource(s) created."
                )
            )
        else:
            return len(created) if created is not None else 0
        return 0

    def _update_resources(
        self,
        resources: Sequence[T_RequestResource],
        loader: ResourceCRUD,
        environment_variable_warning_by_id: dict[Hashable, EnvironmentVariableMissingWarning],
    ) -> int:
        try:
            updated = loader.update(resources)
        except CogniteAPIError as e:
            message = f"Failed to update resource(s). Error: {escape(str(e))}."
            if hint := self._environment_variable_hint(loader.get_ids(resources), environment_variable_warning_by_id):
                message += hint
            raise ResourceUpdateError(message) from e

        return len(updated)

    @staticmethod
    def _environment_variable_hint(
        identifiers: list[Hashable],
        environment_variable_warning_by_id: dict[Hashable, EnvironmentVariableMissingWarning],
    ) -> str:
        if not environment_variable_warning_by_id:
            return ""
        missing_variables: set[str] = set()
        for identifier in identifiers:
            if warning := environment_variable_warning_by_id.get(identifier):
                missing_variables.update(warning.variables)
        if missing_variables:
            variables_str = humanize_collection(missing_variables)
            suffix = "s" if len(missing_variables) > 1 else ""
            return f"\n  {HINT_LEAD_TEXT}This is likely due to missing environment variable{suffix}: {variables_str}"
        return ""

    @staticmethod
    def upload_data(
        loader: DataCRUD,
        state: BuildEnvironment,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> UploadDeployResult:
        prefix = "Would upload" if dry_run else "Uploading"
        print(f"[bold]{prefix} {loader.display_name} files to CDF...[/]")

        datapoints = 0
        file_counts = 0
        for message, file_datapoints in loader.upload(state, dry_run):
            if verbose:
                print(message)
            datapoints += file_datapoints
            file_counts += 1

        if datapoints != 0:
            return DatapointDeployResult(
                loader.display_name, points=datapoints, uploaded=file_counts, item_name=loader.item_name
            )
        else:
            return UploadDeployResult(loader.display_name, uploaded=file_counts, item_name=loader.item_name)
